from qtpy.QtWidgets import QVBoxLayout, QWidget, QGroupBox 
from epicure.laptrack_centroids import LaptrackCentroids
from epicure.laptrack_overlaps import LaptrackOverlaps
from laptrack.data_conversion import convert_split_merge_df_to_napari_graph
from napari.utils import progress
import time
import pandas as pd
import numpy as np
from skimage.measure import regionprops, regionprops_table
from multiprocessing.pool import ThreadPool as Pool
import epicure.Utils as ut
import epicure.epiwidgets as wid

class Tracking(QWidget):
    """
        Handles tracking of cells, track operations with the Tracks layer
    """
    def __init__(self, napari_viewer, epic):
        super().__init__()
        self.viewer = napari_viewer
        self.epicure = epic
        self.graph = None      ## init 
        self.tracklayer = None      ## track layer with information (centroids, labels, tree..)
        self.track_data = None ## keep the updated data, and update the layer only from time to time (slow to do)
        self.tracklayer_name = "Tracks"  ## name of the layer containing tracks
        self.nframes = self.epicure.nframes
        self.properties = ["label", "centroid"]

        layout = QVBoxLayout()
        
        ## Add update track button 
        self.track_update = wid.add_button( "Update tracks display", self.update_track_layer, "Update the Track layer with the changements made since the last update" )
        layout.addWidget(self.track_update)
        

        ## Method specific
        track_method, self.track_choice = wid.list_line( "Tracking method", "Choose the tracking method to use and display its parameter", func=None )
        layout.addWidget(self.track_choice)
        
        self.track_choice.addItem("Laptrack-Centroids")
        self.create_laptrack_centroids()
        layout.addWidget(self.gLapCentroids)
        
        self.track_choice.addItem("Laptrack-Overlaps")
        self.create_laptrack_overlap()
        layout.addWidget(self.gLapOverlap)
        
        self.track_go = wid.add_button( "Track", self.do_tracking, "Launch the tracking with the current parameter. Can take time" )
        layout.addWidget(self.track_go)
        self.setLayout(layout)

        ## General tracking options
        frame_line, self.frame_range, self.range_group = wid.checkgroup_help( "Track only some frames", False, "Option to track only a given range of frames", None ) 
        self.frame_range.clicked.connect( self.show_frame_range )
        range_layout = QVBoxLayout()
        ntrack, self.start_frame = wid.ranged_value_line( "Track from frame:", 0, self.nframes-1, 1, 0, "Set first frame to begin tracking" )
        range_layout.addLayout(ntrack)
        
        entrack, self.end_frame = wid.ranged_value_line( "Until frame:", 1, self.nframes-1, 1, self.nframes-1, "Set the last frame unitl which to track" )
        range_layout.addLayout(entrack)
        self.start_frame.valueChanged.connect( self.changed_start )
        self.end_frame.valueChanged.connect( self.changed_end )
        
        self.range_group.setLayout( range_layout )
        layout.addWidget( self.frame_range )
        layout.addWidget( self.range_group )
        
        self.show_frame_range()
        self.show_trackoptions()
        self.track_choice.currentIndexChanged.connect(self.show_trackoptions)
        

    def show_frame_range( self ):
        """ Show/Hide frame range options """
        self.range_group.setVisible( self.frame_range.isChecked() )
        
    #### settings

    def get_current_settings( self ):
        """ Get current settings to save as preferences """
        settings = {}
        settings["Track method"] = self.track_choice.currentText() 
        settings["Add feat"] = self.check_penalties.isChecked()
        settings["Max distance"] = self.max_dist.text()
        settings["Splitting cost"] = self.splitting_cost.text()
        settings["Merging cutoff"] = self.merging_cost.text()
        settings["Min IOU"] = self.min_iou.text()
        settings["Over split"] = self.split_cost.text()
        settings["Over merge"] = self.merg_cost.text()
        return settings

    def apply_settings( self, settings ):
        """ Set the parameters/current display from the prefered settings """
        for setty, val in settings.items():
            if setty == "Track method":
                self.track_choice.setCurrentText( val )
            if setty == "Add feat":
                self.check_penalties.setChecked( val )
            if setty == "Max distance":
                self.max_dist.setText( val )
            if setty == "Splitting cost":
                self.splitting_cost.setText( val )
            if setty == "Merging cutoff":
                self.merging_cost.setText( val )
            if setty == "Min IOU":
                self.min_iou.setText( val )
            if setty == "Over split":
                self.split_cost.setText( val )
            if setty == "Over merge":
                self.merg_cost.setText( val )
            
    ##########################################
    #### Tracks layer and function

    def reset( self ):
        """ Reset Tracks layer and data """
        self.graph = None
        self.track_data = None
        ut.remove_layer( self.viewer, "Tracks" )

    def init_tracks(self):
        """ Add a track layer with the new tracks """
        track_table, track_prop = self.create_tracks()
        
        ## plot tracks
        if len(track_table) > 0:
            self.viewer.add_tracks(
                track_table,
                graph=self.graph, 
                name=self.tracklayer_name,
                properties = track_prop,)
            self.viewer.layers[self.tracklayer_name].visible=True
            self.viewer.layers[self.tracklayer_name].color_by="track_id"
            ut.set_active_layer(self.viewer, "Segmentation")
            self.tracklayer = self.viewer.layers[self.tracklayer_name]
            self.track_data = self.tracklayer.data
            #self.track.display_id = True
            self.color_tracks_as_labels()

    def color_tracks_as_labels(self):
        """ Color the tracks the same as the label layer """
        ## must color it manually by getting the Label layer colors for each track_id
        cols = np.zeros((len(self.tracklayer.data[:,0]),4))
        for i, tr in enumerate(self.tracklayer.data[:,0]):
            cols[i] = (self.epicure.seglayer.get_color(tr))
        self.tracklayer._track_colors = cols
        self.tracklayer.events.color_by()
    
    def color_tracks_by_lineage(self):
        """ Color the tracks by their lineage (daughters same colors as parents) """
        ## must color it manually by getting the Label layer colors for each track_id
        cols = np.zeros((len(self.tracklayer.data[:,0]),4))
        for i, tr in enumerate(self.tracklayer.data[:,0]):
            ## find the parent cell,n going up the tree until no more parent
            while tr in self.graph.keys():
                tr = self.graph[tr][0]
            cols[i] = (self.epicure.seglayer.get_color(tr))
        self.tracklayer._track_colors = cols
        self.tracklayer.events.color_by()

    def replace_tracks(self, track_df):
        """ Replace all tracks based on the dataframe """
        track_table, track_prop = self.build_tracks(track_df)
        self.tracklayer.data = track_table
        self.track_data = self.tracklayer.data
        self.tracklayer.properties = track_prop
        self.tracklayer.refresh()
        self.color_tracks_as_labels()

    def reset_tracks(self):
        """ Reset tracks and reload them from labels """
        ut.remove_layer(self.viewer, "Tracks")
        self.init_tracks()

    def update_track_layer(self):
        """ Update the track layer (slow) """
        self.viewer.window._status_bar._toggle_activity_dock(True)
        progress_bar = progress(total=1)
        progress_bar.set_description( "Updating track layer" )
        self.tracklayer.data = self.track_data
        progress_bar.close()
        self.color_tracks_as_labels()
        self.viewer.window._status_bar._toggle_activity_dock(False)

    def measure_track_features(self, track_id):
        """ Measure features (length, total displacement...) of given track """
        features = {}
        track = self.get_track_data( track_id )
        start = int(np.min(track[:,1]))
        end = int(np.max(track[:,1]))
        features["Label"] = track_id
        features["TrackDuration"] = end - start + 1
        features["TrackStart"] = start
        features["TrackEnd"] = end
        if (end-start) == 0:
            ## only one frame
            features["TotalDisplacement"] = None
            features["NetDisplacement"] = None
            features["Straightness"] = None
        else:
            features["TotalDisplacement"] = ut.total_distance( track[:,2:4] )
            features["NetDisplacement"] = ut.net_distance( track[:,2:4] )
            if features["TotalDisplacement"] > 0:
                features["Straightness"] = features["NetDisplacement"]/features["TotalDisplacement"]
            else:
                features["Straightness"] = None
        return features

    def measure_features( self, track_id, features ):
        """ Measure features along all the track """
        mask = self.epicure.get_mask( track_id )
        res = {}
        for feat in features:
            res[feat] = []
        for frame in mask:
            props = regionprops( frame )
            if len(props) > 0:
                if "Area" in features:
                    res["Area"].append( props[0].area )
                if "Hull" in features:
                    res["Hull"].append( props[0].area_convex )
                if "Elongation" in features:
                    res["Elongation"].append( props[0].axis_major_length )
                if "Eccentricity" in features:
                    res["Eccentricity"].append( props[0].eccentricity )
                if "Perimeter" in features:
                    res["Perimeter"].append( props[0].perimeter )
                if "Solidity" in features:
                    res["Solidity"].append( props[0].solidity )
        return res

    def measure_specific_feature( self, track_id, featureName ):
        """ Measure some specific feature """
        if featureName == "Similarity":
            import skimage.metrics as imetrics
            movie = self.epicure.get_label_movie( track_id, extend=1.5 )
            sim_scores = []
            for i in range(0, len(movie)-1):
                score = imetrics.normalized_mutual_information( movie[i], movie[i+1] ) 
                sim_scores.append(score)
            return sim_scores

    def measure_labels(self, segimg):
        """ Get the dataframe of the labels in the segmented image """
        resdf = None
        for iframe, frame in progress(enumerate(segimg)):
            frame_table = self.get_one_frame( frame, iframe )
            if resdf is None:
                resdf = pd.DataFrame(frame_table)
            else:
                resdf = pd.concat([resdf, pd.DataFrame(frame_table)])
        return resdf

    def add_track_frame(self, label, frame, centroid, tree=None, group=None):
        """ Add one frame to the track """
        new_frame = np.array([label, frame, centroid[0], centroid[1]])
        self.track_data = np.vstack((self.track_data, new_frame))
            
    def get_track_list(self):
        """ Get list of unique track_ids """
        return np.unique( self.track_data[:,0] )
    
    def get_tracks_on_frame( self, tframe ):
        """ Return list of tracks present on given frame """
        return np.unique( self.track_data[ self.track_data[:,1]==tframe, 0] ) 

    def has_track(self, label):
        """ Test if track label is present """
        return label in self.track_data[:,0]
    
    def has_tracks(self, labels):
        """ Test if track labels are present """
        return np.isin( labels, self.track_data[:,0] )
    
    def nb_points(self):
        """ Number of points in the tracks """
        return self.track_data.shape[0]

    def nb_tracks(self):
        """ Return number of tracks """
        #return self.track._manager.__len__()
        return len(self.get_track_list())

    def gaped_track(self, track_id):
        """ Check if there is a gap (missing frame) in a track """
        indexes = self.get_track_indexes(track_id)
        if len(indexes) <= 0:
            return False
        track_frames = self.track_data[indexes,1]
        return ((np.max(track_frames)-np.min(track_frames)+1) > len(track_frames) )

    def gap_frames(self, track_id):
        """ Returns the frame(s) at which the gap(s) are """
        track_frames = self.get_track_column( track_id, "frame" )
        gaps = []
        if len( track_frames ) > 0:
            min_frame = int( np.min(track_frames) )
            max_frame = int( np.max(track_frames) )
            gaps = np.setdiff1d( np.arange(min_frame+1, max_frame), track_frames ).tolist()
            if len(gaps) > 0:
                gaps.sort()
        return gaps
            
    def check_gap(self, tracks=None, verbose=None):
        """ Check if there is a track with a gap, flag it if yes """
        if tracks is None:
            tracks = self.get_track_list()
        gaped = []
        for track in tracks:
            if self.gaped_track( track ):
                gaped.append(track)
        if verbose is None:
            verbose = self.epicure.verbose
        if verbose > 0 and len(gaped)>0:
            ut.show_warning("Gap in track(s) "+str(gaped)+"\n"
            +"Consider doing sanity_check in Editing onglet to fix it")
        return gaped

    def get_track_indexes(self, track_id):
        """ Get indexes of track_id tracks position in the arrays """
        if isinstance( track_id,  int ):
            return (np.flatnonzero( self.track_data[:,0] == track_id ) )
        return (np.flatnonzero( np.isin( self.track_data[:,0], track_id ) ) )

    def get_track_indexes_from_frame(self, track_id, frame):
        """ Get indexes of track_id tracks position in the arrays from the given frame """
        if type(track_id) == int:
            return (np.argwhere( (self.track_data[:,0] == track_id)*(self.track_data[:,1]>= frame) )).flatten()
        return (np.argwhere( np.isin( self.track_data[:,0], track_id )*(self.track_data[:,1]>= frame) )).flatten()

    def get_index(self, track_id, frame ):
        """ Get index of track_id at given frame """
        if np.isscalar(track_id):
            track_id = [track_id]
        return np.argwhere( (np.isin(self.track_data[:,0], track_id))*(self.track_data[:,1] == frame) )

    def get_small_tracks(self, max_length=1):
        """ Get tracks smaller than the given threshold """
        labels = []
        lengths = []
        positions = []
        for lab in self.get_track_list():
            indexes = self.get_track_indexes(lab)
            length = len(indexes)
            if length <= max_length:
                pos = self.mean_position( indexes, only_first=False )
                labels.append(lab)
                lengths.append(length)
                positions.append(pos)
        return labels, lengths, positions

    def get_track_data(self, track_id):
        """ Return the data of track track_id """
        indexes = self.get_track_indexes( track_id )
        track = self.track_data[indexes,]
        return track
    
    def get_track_column( self, track_id, column ):
        """ Return the chosen column (frame, x, y, label) of track track_id """
        indexes = self.get_track_indexes( track_id )
        if column == "frame":
            i = 1
        if column == "label":
            i = 0
        if column == "pos":
            i = [2,3]
        track = self.track_data[indexes, i]
        return track

    def get_frame_data( self, track_id, ind ):
        """ Get ind-th data of track track_id """
        track = self.get_track_data( track_id )
        return track[ind]

    def get_position( self, track_id, frame ):
        """ Get position of the track at given frame """
        ind = self.get_index( track_id, frame )
        return [int(self.track_data[ind, 2]), int(self.track_data[ind,3])]

    def mean_position(self, indexes, only_first=False):
        """ Mean positions of tracks at indexes """
        if len(indexes) <= 0:
            return None
        track = self.track_data[indexes,]
        ## keep only the first frame of the tracks
        if only_first:
            min_frame = np.min(track[:,1])
            track = track[track[:,1]==min_frame,]
        return ( int(np.mean(track[:,1])), int(np.mean(track[:,2])), int(np.mean(track[:,3])) )

    def get_first_frame(self, track_id):
        """ Returns first frame where track_id is present """
        track = self.get_track_data( track_id )
        if len(track) <= 0:
            return None
        return int( np.min(track[:,1]) )

    
    def get_last_frame(self, track_id):
        """ Returns last frame where track_id is present """
        track = self.get_track_data( track_id )
        return int(np.max(track[:,1]))
    
    def get_extreme_frames(self, track_id):
        """ Returns the first and last frames where track_id is present """
        track = self.get_track_data( track_id )
        if track.shape[0] > 0:
            return (int(np.min(track[:,1])), int(np.max(track[:,1])) )
        return None, None

    def get_mean_position(self, track_id, only_first=False):
        """ Get mean position of the track """
        indexes = self.get_track_indexes( track_id )
        return self.mean_position( indexes, only_first )

    def update_centroid(self, track_id, frame, ind=None, cx=None, cy=None):
        """ Update track at given frame """
        if ind is None:
            ind = self.get_index( track_id, frame )
        if cx is None:
            prop = ut.getPropLabel( self.epicure.seg[frame], track_id )
            self.track_data[ind, 2:4] = prop.centroid[1]
        else:
            self.track_data[ind, 2] = cx
            self.track_data[ind, 3] = cy

    def swap_frame_id(self, tid, otid, frame):
        """ Swap the ids of two tracks at frame """
        ind = int(self.get_index(tid, frame))
        oind = int(self.get_index(otid, frame))
        ## check if one of the label is an extreme of a track and potentially in the graph
        for track_index in [tid, otid]:
            min_frame, max_frame = self.get_extreme_frames( track_index )
            if (min_frame == frame) or (max_frame == frame):
                self.update_graph( track_index, frame )
        self.track_data[[ind, oind],0] = [otid, tid]

    def update_track_on_frame(self, track_ids, frame):
        """ Update (add or modify) tracks at given frame """
        frame_table = regionprops_table( np.where(np.isin(self.epicure.seg[frame], track_ids), self.epicure.seg[frame], 0), properties=self.properties )
        for x, y, tid in zip(frame_table["centroid-0"], frame_table["centroid-1"], frame_table["label"]):
            index = self.get_index(tid, frame)
            if len(index) > 0:
                self.update_centroid( tid, frame, index, int(x), int(y) )
            else:
                cur_cell = np.array( [tid, frame, int(x), int(y)] )
                cur_cell = np.expand_dims(cur_cell, axis=0)
                self.track_data = np.append(self.track_data, cur_cell, axis=0)
    
    def add_one_frame(self, track_ids, frame, refresh=True):
        """ Add one frame from track """
        for tid in track_ids:
            frame_table = regionprops_table(np.uint8(self.epicure.seg[frame]==tid), properties=self.properties)
            cur_cell = np.array( [tid, frame, int(frame_table["centroid-0"]), int(frame_table["centroid-1"])], dtype=np.uint32 )
            cur_cell = np.expand_dims(cur_cell, axis=0)
            self.track_data = np.append(self.track_data, cur_cell, axis=0)

    def remove_one_frame( self, track_id, frame, handle_gaps=False, refresh=True ):
        """ 
        Remove one frame from track(s) 
        """
        inds = self.get_index( track_id, frame )
        if np.isscalar(track_id):
            track_id = [track_id]
        check_for_gaps = False
        for tid in track_id:
            ## removed frame is in the extremity of a track, can be in the graph
            first_frame, last_frame = self.get_extreme_frames( tid )
            if first_frame is None:
                continue
            if (first_frame == frame) or (last_frame == frame):
                self.update_graph( tid, frame )
            else:
                check_for_gaps = True
        self.track_data = np.delete( self.track_data, inds, axis=0 )
        ## gaps might have been created in the tracks, for now doesn't allow it so split the tracks
        if handle_gaps and check_for_gaps:
            gaped = self.check_gap( track_id, verbose=0 )
            if len(gaped) > 0:
                self.epicure.fix_gaps( gaped )
        
    def get_current_value(self, track_id, frame):
        ind = self.get_index( track_id, frame )
        centx = int(self.track_data[ind, 2])
        centy = int(self.track_data[ind, 3])
        return self.epicure.seg[frame, centx, centy]

    def update_graph(self, track_id, frame):
        """ Update graph if deleted label was linked at that frame, assume keys are unique """
        start_time = time.time()
        if self.graph is not None:
            ## handles current node is last of his branch
            parents = self.last_in_graph( track_id, frame )
            if len(parents) > 0:
                current_label = self.get_current_value( track_id, frame )
                if current_label == 0:
                    for parent in parents:
                        del self.graph[parent]
                else:
                    for parent in parents:
                        self.update_child( parent, track_id, current_label )
            ## handles when current track is first frame of a division
            if self.first_in_graph( track_id, frame ):
                current_label = self.get_current_value( track_id, frame )
                if current_label == 0:
                    del self.graph[track_id]
                else:
                    self.update_key( track_id, current_label ) 
        if self.epicure.verbose > 1:
            ut.show_duration( start_time, header="Graph updated in " )

    def update_child(self, parent, prev_key, new_key):
        """ Change the value of a key in the graph """
        vals = self.graph[parent]
        self.graph[parent] = []
        for val in vals:
            if val == prev_key:
                self.graph[parent].append(new_key)
            else:
                self.graph[parent].append(val)

    def update_key(self, prev_key, new_key):
        """ Change the value of a key in the graph """
        self.graph[new_key] = self.graph.pop(prev_key)

    def is_parent( self, cur_id ):
        """ Return if the current id is in the graph (as a parent, so in values) """
        if self.graph is None:
            return False
        return any( cur_id in vals for vals in self.graph.values() )

    def add_division( self, childa, childb, parent ):
        """ Add info of a division to the graph of divisions/merges """
        if self.graph is None:
            self.graph = {}
        self.graph[childa] = parent
        self.graph[childb] = parent

    def remove_division( self, parent ):
        """ Remove a division event from the graph """
        keys = list(self.graph.keys())
        for key in keys:
            vals = self.graph[key]
            if isinstance(vals, list):
                if vals[0] == parent:
                    del self.graph[key]
            else:
                if vals == parent:
                    del self.graph[key]

    def last_in_graph(self, track_id, frame):
        """ Check if given label and frame is the last of a branch, in the graph """
        parents = []
        for key, vals in self.graph.items():
            if not isinstance(vals, list):
                vals = [vals]
            for val in vals:
                if val == track_id:
                    last_frame = self.get_last_frame( val )
                    if last_frame == frame:
                        parents.append(key)
        return parents

    def first_in_graph(self, track_id, frame):
        """ Check if the given label and frame is the first in the branch so the node in the graph """
        if track_id in self.graph.keys():
            first_frame = self.get_first_frame(track_id)
            if first_frame == frame:
                return True
        return False

    def remove_tracks(self, track_ids):
        """ Remove track with given id """
        inds = self.get_track_indexes(track_ids)
        self.track_data = np.delete(self.track_data, inds, axis=0)
        self.remove_ids_from_graph( track_ids )
    
    def remove_ids_from_graph( self, track_ids ):
        """ Remove all ids from the graph """
        track_ids_set = set( track_ids )
        self.graph = {
            key: vals for key, vals in self.graph.items()
            if (key not in track_ids_set) and ( not any( val in track_ids_set for val in (vals if isinstance(vals, list) else [vals])) )
        }
        """
        for tid in track_ids:
            self.graph.pop( tid, None )
        
        keys = self.graph.keys()
        for key in keys:
            kgr = self.graph[ key ]
            if not isinstance( kgr, list ):
                kgr = [kgr]
            for val in kgr:
                if val in track_ids:
                    del self.graph[key]
                    continue
                """
       
    def build_tracks(self, track_df):
        """ Create tracks from dataframe (after tracking) """
        track = track_df[["track_id", "frame", "centroid-0", "centroid-1"]]
        #frame_prop = frame_table[["tree_id", "label", "nframes", "group"]]
        return np.array(track, int), None #dict(frame_prop)

    def create_tracks(self):
        """ Create tracks from labels (without tracking) """
        properties = ['label', 'centroid']
        track_table = np.empty((0,4), int)
        if self.epicure.nframes >= 0:
            for iframe, frame in progress(enumerate(self.epicure.seg)):
                frame_track, frame_prop = self.get_one_frame( frame, iframe )
                track_table = np.vstack((track_table, frame_track))
        
        return track_table, None # track_prop

    def add_track_features(self, labels):
        """ Add features specific to tracks (eg nframes) """
        nframes = np.zeros(len(labels), int)
        if self.epicure.verbose > 2:
            print("REPLACE BY COUNT METHOD")
        for track_id in np.unique(labels):
            cur_track = np.argwhere(labels == track_id)
            nframes[ list(cur_track) ] = len(cur_track)
        return nframes
    
    def get_one_frame(self, seg, frame):
        """ Get the regionprops results into the dataframe """
        frame_table = regionprops_table(seg, properties=self.properties)
        ndata = len(frame_table["label"])
        if frame is not None:
            frame_table["frame"] = np.repeat(frame, ndata)
        frame_table["track_id"] = frame_table["label"]
        #frame_table["tree_id"] = [0]*ndata    ##### @Tochange to read div/merge infos ?
        #frame_table["group"] = [0]*ndata    ##### @Tochange to read div/merge infos ?
        #frame_table["nframes"] = [0]*ndata
        frame_table = pd.DataFrame(frame_table)
        frame_track = frame_table[["track_id", "frame", "centroid-0", "centroid-1"]]
        #frame_prop = frame_table[["tree_id", "label", "nframes", "group"]]
        return np.array(frame_track, int), None #dict(frame_prop)


    ##########################################
    #### Tracking functions

    def changed_start(self, i):
        """ Ensures that end frame > start frame """
        if i > self.end_frame.value():
            self.end_frame.setValue(i+1)

    def changed_end(self, i):
        if i < self.start_frame.value():
            self.start_frame.setValue(i-1)

    def find_parents(self, labels, twoframes):
        """ Find in the first frame the parents of labels from second frame """
        
        if self.track_choice.currentText() == "Laptrack-Centroids":
            return self.laptrack_centroids_twoframes(labels, twoframes)
        
        if self.track_choice.currentText() == "Laptrack-Overlaps":
            return self.laptrack_overlaps_twoframes(labels, twoframes)
        

    def do_tracking(self):
        """ Start the tracking with the selected options """
        if self.frame_range.isChecked():
            start = self.start_frame.value()
            end = self.end_frame.value()
        else:
            start = 0
            end = self.nframes-1
        start_time = time.time()
        self.viewer.window._status_bar._toggle_activity_dock(True)
        self.epicure.inspecting.reset_all_events()
        
        if self.track_choice.currentText() == "Laptrack-Centroids":
            if self.epicure.verbose > 1:
                print("Starting track with Laptrack-Centroids")
            self.laptrack_centroids( start, end )
            self.epicure.tracked = 1
        if self.track_choice.currentText() == "Laptrack-Overlaps":
            if self.epicure.verbose > 1:
                print("Starting track with Laptrack-Centroids")
            self.laptrack_overlaps( start, end )
            self.epicure.tracked = 1
        
        self.epicure.finish_update(contour=2)
        #self.epicure.reset_free_label()
        self.viewer.window._status_bar._toggle_activity_dock(False)
        if self.epicure.verbose > 0:
            ut.show_duration( start_time, header="Tracking done in " )

    def show_trackoptions(self):
        self.gLapCentroids.setVisible(self.track_choice.currentText() == "Laptrack-Centroids")
        self.gLapOverlap.setVisible(self.track_choice.currentText() == "Laptrack-Overlaps")

    def relabel_nonunique_labels(self, track_df):
        """ After tracking, some track can be splitted and get same label, fix that """
        inittids = np.unique(track_df["track_id"])
        labtracks = []
        saved_data = np.copy(self.epicure.seglayer.data)
        mframes = []
        tids = []
        used = np.unique( saved_data )
        all_labels = np.unique(track_df["label"])
        for tid in inittids:
            cdf = track_df[track_df["track_id"]==tid]
            #print(cdf)
            min_frame = np.min( cdf["frame"] )
            #labtrack = int( cdf["label"][cdf["frame"]==min_frame] )
            for lab in np.unique(cdf["label"]):
                labtracks.append(lab)
                mframes.append( min_frame )
                tids.append(tid)
        if len(labtracks) != len(np.unique(labtracks)):
            ## some labels are present several times
            used = used.tolist()
            for lab in all_labels :
                indexes = list(np.where(np.array(labtracks)==lab)[0])
                if len(indexes)>1:
                    minframes = [mframes[indy] for indy in range(len(labtracks)) if labtracks[indy]==lab]
                    indmin = indexes[ np.argmin( minframes ) ]
                    ## for the other(s), change the label
                    newvals = ut.get_free_labels( used, len(indexes) )
                    used = used + newvals
                    for i, ind in enumerate(indexes):
                        if ind != indmin:
                            tid = tids[ind]
                            newval = newvals[i]
                            track_df.loc[ (track_df["track_id"]==tid)  & (track_df["label"]==lab) , "label"] = newval
                            for frame in track_df["frame"][(track_df["track_id"]==tid) & (track_df["label"]==newval)]:
                                mask = (saved_data[frame]==lab)
                                self.epicure.seglayer.data[frame][mask] = newval
        

    def relabel_trackids(self, track_df, splitdf, mergedf):
        """ Change the trackids to take the first label of each track """
        replace_map = dict()
        new_trackids = track_df['track_id'].copy() 
        new_splitdf = splitdf.copy()
        new_mergedf = mergedf.copy()
        for tid in np.unique(track_df['track_id']):
            ctrack_df = track_df[track_df['track_id']==tid]
            newval = int( ctrack_df.loc[ ctrack_df["frame"]==np.min(ctrack_df["frame"]), "label" ].iloc[0] )
            ## have to replace if different
            if tid != newval:
                new_trackids.loc[ track_df['track_id']==tid ] = newval
                if not new_splitdf.empty:
                    new_splitdf.loc[ splitdf["parent_track_id"]==tid, "parent_track_id" ] = newval
                    new_splitdf.loc[ splitdf["child_track_id"]==tid, "child_track_id" ] = newval
                if not new_mergedf.empty:
                    new_mergedf.loc[ mergedf["parent_track_id"]==tid, "parent_track_id" ] = newval
                    new_mergedf.loc[ mergedf["child_track_id"]==tid, "child_track_id" ] = newval
        return new_trackids, new_splitdf, new_mergedf

    def change_labels(self, track_df):
        """ Change the labels at each frame according to tracks """
        frames = track_df["frame"]
        ## change the other ones
        for frame in np.unique(frames):
            self.change_frame_labels(frame, track_df)

    def change_frame_labels(self, frame, track_df):
        """ Change the labels at given frame according to tracks """
        frame_df = track_df[track_df["frame"]==frame]
        #coordinates = frame_df[['centroid-0', 'centroid-1']].astype(int).values
        track_ids = frame_df['track_id'].astype(int).values
        old_labels = frame_df["label"].astype(int).values
        seglayer = np.copy(self.epicure.seglayer.data[frame])
        for i, lab in enumerate(old_labels):
            mask = (seglayer==lab)
            self.epicure.seglayer.data[frame][mask] = track_ids[i]

    def label_to_dataframe( self, labimg, frame ):
        """ from label, get dataframe of centroids with properties """
        df = pd.DataFrame(regionprops_table(labimg, properties=self.region_properties))
        df["frame"] = frame
        return df
    
    def labels_to_centroids(self, start_frame, end_frame, locked=True):
        """ Get centroids of each cell in dataframe """
        regionprops = []
        #if self.epicure.process_parallel:
        #    with Pool(self.epicure.nparallel) as pool:
        #        regionprops = pool.map( self.label_frame_todf, range(start_frame, end_frame+1) )
        #    regionprops_df = pd.concat(regionprops)
        #else:
        for frame in range(start_frame, end_frame+1):
            df = self.label_frame_todf(frame)
            regionprops.append(df)
        regionprops_df = pd.concat(regionprops)
        return regionprops_df
    
    def labels_ready(self, start_frame, end_frame, locked=True):
        """ Get labels of unlocked cells to track """
        res_labels = []
        if self.epicure.process_parallel:
            with Pool(self.epicure.nparallel) as pool:
                labels = pool.map( self.current_label_frame, range(start_frame, end_frame+1) )
            res_labels = np.array(labels)
        else:
            for frame in range(start_frame, end_frame+1):
                labels = self.current_label_frame(frame)
                res_labels.append(labels)
            res_labels = np.array(res_labels)
        return res_labels
    
    def label_frame_todf( self, frame ):
        """ For current frame, get label frame image then dataframe of centroids """
        clabel = self.current_label_frame(frame)
        return self.label_to_dataframe( clabel, frame )
    
    def current_label_frame( self, frame ):
        """ For current frame, get label frame image """
        clabel = None
        #if self.track_only_in_roi.isChecked():
        #    clabel = self.epicure.only_current_roi(frame)
        if clabel is None:
            clabel = self.epicure.seg[frame]
        return clabel

    def after_tracking(self, track_df, split_df, merge_df, progress_bar, indprogress):
        """ Steps after tracking: get/show the graph from the track_df """
        graph = None
        progress_bar.set_description( "Update labels and tracks" )
        ## shift all by 1 so that doesn't start at 0
        if len(split_df) > 0:
            split_df[["parent_track_id"]] += 1
            split_df[["child_track_id"]] += 1
        if len(merge_df) > 0:
            merge_df[["parent_track_id"]] += 1
            merge_df[["child_track_id"]] += 1
        track_df[["track_id"]] += 1
       
        ## relabel if some track have the same label
        self.relabel_nonunique_labels(track_df)
        ## relabel track ids so that they are equal to the first label of the track
        newtids, split_df, merge_df = self.relabel_trackids(track_df, split_df, merge_df)
        track_df["track_id"] = newtids
        self.change_labels(track_df)

        # create graph of division/merging
        self.graph = convert_split_merge_df_to_napari_graph(split_df, merge_df)

        progress_bar.update(indprogress+1)
        
        ## update display if active
        self.replace_tracks(track_df)

        progress_bar.update(indprogress+2)
        ## update the list of events, or others 
        self.epicure.updates_after_tracking()
        progress_bar.update(indprogress+3)
        return graph


    
############ Laptrack centroids option
    
    def create_laptrack_centroids(self):
        """ GUI of the laptrack option """
        self.gLapCentroids = QGroupBox("Laptrack-Centroids")
        glap_layout = QVBoxLayout()
        mdist, self.max_dist = wid.value_line( "Max distance", "15.0", "Maximal distance between two labels in consecutive frames to link them (in pixels)" )
        glap_layout.addLayout(mdist)
        ## splitting ~ cell division
        scost, self.splitting_cost = wid.value_line( "Splitting cutoff", "1", "Weight to split a track in two (increasing it favors division)" )
        glap_layout.addLayout(scost)
        ## merging ~ error ?
        mcost, self.merging_cost = wid.value_line( "Merging cutoff", "1", "Weight to merge to labels together" )
        glap_layout.addLayout(mcost)

        add_feat, self.check_penalties, self.bpenalties = wid.checkgroup_help( "Add features cost", True, "Add cell features in the tracking calculation", None )
        self.create_penalties()
        glap_layout.addWidget(self.check_penalties)
        glap_layout.addWidget(self.bpenalties)
        self.gLapCentroids.setLayout(glap_layout)

    def show_penalties(self):
        self.bpenalties.setVisible(not self.bpenalties.isVisible())

    def create_penalties(self):
        pen_layout = QVBoxLayout()
        areaCost, self.area_cost = wid.value_line( "Area difference", "2", "Weight of the difference of area between two labels to link them (0 to ignore)" )
        pen_layout.addLayout(areaCost)
        solidCost, self.solidity_cost = wid.value_line( "Solidity difference", "0", "Weight of the difference of solidity between two labels to link them (0 to ignore)" )
        pen_layout.addLayout(solidCost)
        self.bpenalties.setLayout(pen_layout)

    def laptrack_centroids_twoframes(self, labels, twoframes):
        """ Perform tracking of two frames with current parameters """
        laptrack = LaptrackCentroids(self, self.epicure)
        laptrack.max_distance = float(self.max_dist.text())
        self.region_properties = ["label", "centroid"]
        if self.check_penalties.isChecked():
            self.region_properties.append("area")
            self.region_properties.append("solidity")
            laptrack.penal_area = float(self.area_cost.text())
            laptrack.penal_solidity = float(self.solidity_cost.text())
        laptrack.set_region_properties(with_extra=self.check_penalties.isChecked())
            
        df = self.twoframes_centroid(twoframes)
        if set(np.unique(df["label"])) == set(labels):
            ## no other labels
            return [None]*len(labels) 
        laptrack.splitting_cost = False ## disable splitting option
        laptrack.merging_cost = False ## disable merging option
        parent_labels = laptrack.twoframes_track(df, labels)
        return parent_labels
    
    def twoframes_centroid(self, img):
        """ Get centroids of first frame only """
        df0 = self.label_to_dataframe( img[0], 0 )
        df1 = self.label_to_dataframe( img[1], 1 )
        return pd.concat([df0, df1])
    
    def laptrack_centroids(self, start, end):
        """ Perform track with laptrack option and chosen parameters """
        ## Laptrack tracker
        laptrack = LaptrackCentroids(self, self.epicure)
        laptrack.max_distance = float(self.max_dist.text())
        laptrack.splitting_cost = float(self.splitting_cost.text())
        laptrack.merging_cost = float(self.merging_cost.text())
        self.region_properties = ["label", "centroid"]
        if self.check_penalties.isChecked():
            self.region_properties.append("area")
            self.region_properties.append("solidity")
            laptrack.penal_area = float(self.area_cost.text())
            laptrack.penal_solidity = float(self.solidity_cost.text())
        laptrack.set_region_properties(with_extra=self.check_penalties.isChecked())

        progress_bar = progress(total=7)
        progress_bar.set_description( "Prepare tracking" )
        if self.epicure.verbose > 1:
            print("Convert labels to centroids: use track info ?")
        df = self.labels_to_centroids(start, end, locked=True)
        progress_bar.update(1)
        if self.epicure.verbose > 1:
            print("GO tracking")
        progress_bar.set_description( "Do tracking with LapTrack Centroids" )
        track_df, split_df, merge_df = laptrack.track_centroids(df)
        progress_bar.update(2)
        if self.epicure.verbose > 1:
            print("After tracking, update everything")
        self.after_tracking(track_df, split_df, merge_df, progress_bar, 2)
        progress_bar.update(6)
        progress_bar.close()
    
############ Laptrack overlap option

    def create_laptrack_overlap(self):
        """ GUI of the laptrack overlap option """
        self.gLapOverlap = QGroupBox("Laptrack-Overlaps")
        glap_layout = QVBoxLayout()
        miou, self.min_iou = wid.value_line( "Min IOU", "0.1", "Minimum Intersection Over Union score to link to labels together" )
        glap_layout.addLayout(miou)
        
        scost, self.split_cost = wid.value_line( "Splitting cost", "0.2", "Weight of linking a parent label with two labels (increasing it for more divisions)" )
        glap_layout.addLayout(scost)
        
        mcost, self.merg_cost = wid.value_line( "Merging cost", "0.2", "Weight of merging two parent labels into one" )
        glap_layout.addLayout(mcost)

        self.gLapOverlap.setLayout(glap_layout)

    def laptrack_overlaps(self, start, end):
        """ Perform track with laptrack overlap option and chosen parameters """
        ## Laptrack tracker
        laptrack = LaptrackOverlaps(self, self.epicure)
        miniou = float(self.min_iou.text())
        if miniou >= 1.0:
            miniou = 1.0
        laptrack.cost_cutoff = 1.0 - miniou
        laptrack.splitting_cost = float(self.split_cost.text())
        laptrack.merging_cost = float(self.merg_cost.text())
        self.region_properties = ["label", "centroid"]

        progress_bar = progress(total=6)
        progress_bar.set_description( "Prepare tracking" )
        labels = self.labels_ready(start, end, locked=True)
        progress_bar.update(1)
        progress_bar.set_description( "Do tracking with LapTrack Overlaps" )
        track_df, split_df, merge_df = laptrack.track_overlaps(labels)
        progress_bar.update(2)
        
        ## get dataframe of coordinates to create the graph 
        df = self.labels_to_centroids(start, end, locked=True)
        progress_bar.update(3)
        coordinate_df = df.set_index(["frame", "label"])
        tdf = track_df.set_index(["frame", "label"])
        track_df2 = pd.merge( tdf, coordinate_df, right_index=True, left_index=True).reset_index()
        self.after_tracking(track_df2, split_df, merge_df, progress_bar, 3)
        progress_bar.update(6)
        progress_bar.close()
    
    def laptrack_overlaps_twoframes(self, labels, twoframes):
        """ Perform tracking of two frames with current parameters """
        laptrack = LaptrackOverlaps(self, self.epicure)
        miniou = float(self.min_iou.text())
        if miniou >= 1.0:
            miniou = 1.0
        laptrack.cost_cutoff = 1.0 - miniou
        self.region_properties = ["label", "centroid"]

        laptrack.splitting_cost = False ## disable splitting option
        laptrack.merging_cost = False ## disable merging option
        parent_labels = laptrack.twoframes_track(twoframes, labels)
        return parent_labels





import epicure.Utils as ut
from qtpy.QtWidgets import QPushButton, QCheckBox, QHBoxLayout, QLabel, QLineEdit, QComboBox, QSpinBox, QSlider, QGroupBox
from qtpy.QtCore import Qt

def help_button( link, description="", display_settings=None ):
    """ Create a new Help button with given parameter """
    def show_doc():
        """ Open documentation page """
        ut.show_documentation_page( link )

    help_btn = QPushButton( "help" )
    if description == "":
        help_btn.setToolTip( "Open EpiCure documentation" )
        help_btn.setStatusTip( "Open EpiCure documentation" )
    else:
        help_btn.setToolTip( description )
        help_btn.setStatusTip( description )
    help_btn.clicked.connect( show_doc )
    if display_settings is not None:
        if "Help button" in display_settings:
            color = display_settings["Help button"]
            help_btn.setStyleSheet( 'QPushButton {background-color: '+color+'}' )
    return help_btn

def checkgroup_help( name, checked, descr, help_link, display_settings=None, groupnb=None ):
    """ Create a group that can be show/hide with checkbox and an help button """
    group = QGroupBox( name )
    chbox = QCheckBox( text=name )

    ## set group and checkbox to the same specific color
    if (groupnb is not None) and (display_settings is not None):
        if groupnb in display_settings:
            color = display_settings[groupnb]
            group.setStyleSheet( 'QGroupBox {background-color: '+color+'}' )
            chbox.setStyleSheet( 'QCheckBox::indicator {background-color: '+color+'}' )
    
    def show_hide():
        group.setVisible( chbox.isChecked() )

    line = QHBoxLayout()
    ## create checkbox
    chbox.setToolTip( descr )
    line.addWidget( chbox )
    chbox.stateChanged.connect( show_hide )
    chbox.setChecked( checked )
    ## create button
    if help_link is not None:
        help_btn = help_button( help_link, "", display_settings )
        line.addWidget( help_btn )
    return line, chbox, group

def checkhelp_line( checkbox_name, checked, checkfunc, check_descr, help_link, display_settings=None, help_descr="" ):
    """ Create a layout line with a checkbox associated with help button """
    line = QHBoxLayout()
    ## create checkbox
    chbox = QCheckBox( text=checkbox_name )
    chbox.setToolTip( check_descr )
    line.addWidget( chbox )
    if checkfunc is not None:
        chbox.stateChanged.connect( checkfunc )
    chbox.setChecked( checked )
    ## create button
    help_btn = help_button( help_link, help_descr, display_settings )
    line.addWidget( help_btn )
    return line, chbox

def add_check( check, checked, check_func=None, descr="" ):
    """ Add a checkbox with set parameters """
    cbox = QCheckBox( text=check )
    cbox.setToolTip( descr )
    if check_func is not None:
        cbox.stateChanged.connect( check_func )
    cbox.setChecked( checked )
    return cbox

def double_check( checka, checkeda, funca, descra, checkb, checkedb, funcb, descrb ):
    """ Line with two customized checkboxes """
    line = QHBoxLayout()
    check_a = add_check( checka, checkeda, funca, descra )
    check_b = add_check( checkb, checkedb, funcb, descrb )
    line.addWidget( check_a )
    line.addWidget( check_b )
    return line, check_a, check_b

def add_button( btn, btn_func, descr="", color=None ):
    """ Add a button connected to an action when pushed """
    btn = QPushButton( btn )
    if btn_func is not None:
        btn.clicked.connect( btn_func )
    if descr != "":
        btn.setToolTip( descr )
    else:
        btn.setToolTip( "Click to perform action" )
    if color is not None:
        btn.setStyleSheet( 'QPushButton {background-color: '+color+'}' )
    return btn

def double_button( btna, funca, descra, btnb, funcb, descrb ):
    """ Line with two customized buttons """
    line = QHBoxLayout()
    btn_a = add_button( btna, funca, descra )
    btn_b = add_button( btnb, funcb, descrb )
    line.addWidget( btn_a )
    line.addWidget( btn_b )
    return line

def button_parameter_line( btn, btn_func, value, descr_btn="", descr_value="" ):
    """ Create a layout with a button and an editable value associated """
    line = QHBoxLayout()
    ## Action button
    btn = QPushButton( btn )
    btn.clicked.connect( btn_func )
    if descr_btn != "":
        btn.setToolTip( descr_btn )
    line.addWidget( btn )
    ## Value editable
    val = QLineEdit()
    val.setText( value )
    line.addWidget( val )
    if descr_value != "":
        val.setToolTip( descr_value )
    return line, val

def min_button_max( btn, btn_func, min_val, max_val, descr="" ):
    """ Button inside two values (min and max) interfaces """
    line = QHBoxLayout()
    ## left value
    minv = QLineEdit()
    minv.setText( min_val )
    line.addWidget( minv )
    ## button
    btn = QPushButton( btn )
    btn.clicked.connect( btn_func )
    if descr != "":
        btn.setToolTip( descr )
    line.addWidget( btn )
    ## right value
    maxv = QLineEdit()
    maxv.setText( max_val )
    line.addWidget( maxv )
    return line, minv, maxv


def button_check_line( btn, btn_func, check, checked=False, checkfunc=None, descr_btn="", descr_check="", leftbtn=True ):
    """ Create a layout with a button and an assiociated checkbox """
    line = QHBoxLayout()
    ## Action button
    btn = QPushButton( btn )
    btn.clicked.connect( btn_func )
    if descr_btn != "":
        btn.setToolTip( descr_btn )
    ## Value editable
    cbox = QCheckBox( check )
    if descr_check != "":
        cbox.setToolTip( descr_check )
    if checkfunc is not None:
        cbox.stateChanged.connect( checkfunc )
    cbox.setChecked( checked )
    ## button first (left), then checkbox
    if leftbtn:
        line.addWidget( btn )
        line.addWidget( cbox )
    else:
        ## or checkbox first (left), then button
        line.addWidget( cbox )
        line.addWidget( btn )
    return line, cbox

def value_line( label, default_value, descr="" ):
    """ Create a layout line with a value to edit (non editable name + value part ) """
    line = QHBoxLayout()
    ## Value name
    lab = QLabel()
    lab.setText( label )
    line.addWidget( lab )
    if descr != "":
        lab.setToolTip( descr )
    ## Value editable part
    value = QLineEdit()
    value.setText( default_value )
    line.addWidget( value )
    return line, value

def check_value( check, checkfunc=None, checked=False, value="0", descr="", label=None ):
    """ Line with a checkbox and an associated editable parameter """
    line = QHBoxLayout()
    ## add checkbox
    cbox = add_check( check, checked=checked, check_func=checkfunc, descr=descr )
    line.addWidget( cbox )
    ## add eventually a text
    if label is not None:
        lab = QLabel()
        lab.setText( label )
        line.addWidget( lab )
    ## add the editable value
    val = QLineEdit()
    val.setText( value )
    line.addWidget( val )
    return line, cbox, val

def ranged_value_line( label, minval, maxval, step, val, descr="" ):
    """ Create a line with a label and a ranged value (limited between min and max) """
    line = QHBoxLayout()
    ## Add the name of the value
    lab = QLabel()
    lab.setText( label )
    if descr != "":
        lab.setToolTip( descr )
    line.addWidget( lab )
    ## Ranged-value widget
    ranged_val = QSpinBox()
    ranged_val.setMinimum( minval )
    ranged_val.setMaximum( maxval )
    ranged_val.setSingleStep( step ) 
    ranged_val.setValue( val )
    line.addWidget( ranged_val )
    return line, ranged_val

def button_list( btn, func, descr ):
    """ Button associated with a list """
    line = QHBoxLayout()
    ## Button part
    button = add_button( btn, func, descr )
    line.addWidget( button )
    ## list part
    li = QComboBox()
    line.addWidget( li )
    return line, li

def list_line( label, descr="", func=None ):
    """ Create a layout line with a choice list to edit (non editable name + list part ) """
    line = QHBoxLayout()
    ## Value name
    lab = QLabel()
    lab.setText( label )
    line.addWidget( lab )
    if descr != "":
        lab.setToolTip( descr )
        lab.setStatusTip( descr )
    ## Value editable part
    value = QComboBox()
    line.addWidget( value )
    if func is not None:
        value.currentIndexChanged.connect( func )
    return line, value

def slider_line( name, minval, maxval, step, value, slidefunc=None, descr="" ):
    """ Line with a text and a slider """
    line = QHBoxLayout()
    ## add name if any
    if name is not None:
        lab = QLabel()
        lab.setText( name )
        line.addWidget( lab )
    ## add slider
    slider =  QSlider( Qt.Horizontal )
    slider.setMinimum( minval )
    slider.setMaximum( maxval )
    slider.setSingleStep( step )
    slider.setValue( value )
    if slidefunc is not None:
        slider.valueChanged.connect( slidefunc )
    if descr != "":
        slider.setToolTip( descr )
    line.addWidget( slider )
    return line, slider



from qtpy.QtWidgets import QHBoxLayout, QVBoxLayout, QWidget, QGroupBox, QComboBox, QLabel, QCheckBox, QTableWidget, QTableWidgetItem, QGridLayout, QListWidget
from qtpy.QtWidgets import QAbstractItemView as aiv
from qtpy.QtCore import Qt
import pandas as pand
import numpy as np
import roifile
from skimage.morphology import binary_erosion, binary_dilation, disk
from skimage.segmentation import expand_labels
import os, time
import napari
from napari.utils import progress
import epicure.Utils as ut
import epicure.epiwidgets as wid
import matplotlib as mpl
from matplotlib.backends.backend_qt5agg import FigureCanvas
from matplotlib.figure import Figure
import matplotlib.pyplot as plt

try:
    from skimage.graph import RAG
except:
    from skimage.future.graph import RAG  ## older version of scikit-image
    

class Outputing(QWidget):

    def __init__(self, napari_viewer, epic):
        super().__init__()
        self.viewer = napari_viewer
        self.epicure = epic
        self.table = None
        self.table_selection = None
        self.seglayer = self.viewer.layers["Segmentation"]
        self.movlayer = self.viewer.layers["Movie"]
        self.selection_choices = ["All cells", "Only selected cell"]
        self.output_options = ["", "Export to extern plugins", "Export segmentations", "Measure cell features", "Measure track features", "Export events"]
        self.tplots = None
        
        chanlist = ["Movie"]
        if self.epicure.others is not None:
            for chan in self.epicure.others_chanlist:
                chanlist.append( "MovieChannel_"+str(chan) )
        self.cell_features = CellFeatures( chanlist )
        self.event_types = EventTypes() 
        
        all_layout = QVBoxLayout()
        
        self.choose_output = QComboBox()
        all_layout.addWidget(self.choose_output)
        for option in self.output_options:
            self.choose_output.addItem(option)
        self.choose_output.currentIndexChanged.connect(self.show_output_option)
        
        ## Choice of active selection
        layout = QVBoxLayout()
        selection_layout, self.output_mode = wid.list_line( "Apply on", descr="Choose on which cell(s) to do the action", func=None )
        for sel in self.selection_choices:
            self.output_mode.addItem(sel)
        all_layout.addLayout(selection_layout)
       
        ## Choice of interface
        self.export_group = QGroupBox("Export to extern plugins")
        export_layout = QVBoxLayout()
        griot_btn = wid.add_button( "Current frame to Griottes", self.to_griot, "Launch(in new window) Griottes plugin on current frame" )
        export_layout.addWidget(griot_btn)
        ncp_btn = wid.add_button( "Current frame to Cluster-Plotter", self.to_ncp, "Launch (in new window) cluster-plotter plugin on current frame" )
        export_layout.addWidget(ncp_btn)
        self.export_group.setLayout(export_layout)
        all_layout.addWidget(self.export_group)
        
        ## Option to export segmentation results
        self.export_seg_group = QGroupBox(self.output_options[2])
        layout = QVBoxLayout()
        save_line, self.save_choice = wid.button_list( "Save segmentation as", self.save_segmentation, "Save the current segmentation either as ROI, label image or skeleton" ) 
        self.save_choice.addItem( "labels" )
        self.save_choice.addItem( "ROI" )
        self.save_choice.addItem( "skeleton" )
        layout.addLayout( save_line )

        self.export_seg_group.setLayout(layout)
        all_layout.addWidget(self.export_seg_group)
        
        #### Features group
        self.feature_group = QGroupBox(self.output_options[3])
        featlayout = QVBoxLayout()
        
        self.choose_features_btn = wid.add_button( "Choose features", self.choose_features, "Open a window to select the features to measure" )
        featlayout.addWidget(self.choose_features_btn)

        self.feature_table = wid.add_button( "Create features table", self.show_table, "Measure the selected features and display it as a clickable table" )
        featlayout.addWidget(self.feature_table)
        self.featTable = FeaturesTable(self.viewer, self.epicure)
        featlayout.addWidget(self.featTable)
        
        ######## Temporal option  
        self.temp_graph = wid.add_button( "Table to temporal graphs", self.temporal_graphs, "Open a plot interface of measured features temporal evolution" )
        featlayout.addWidget(self.temp_graph)
        self.temp_graph.setEnabled(False)
       
        ######## Drawing option
        featmap, self.show_feature_map = wid.list_line( "Draw feature map:", descr="Add a layer with the cells colored by the selected feature value", func=self.show_feature )
        featlayout.addLayout(featmap)
        orienbtn = wid.add_button( "Draw cell orientation", self.draw_orientation, "Add a layer with each cell main axis orientation and length " )
        featlayout.addWidget( orienbtn )

        self.save_table = wid.add_button( "Save features table", self.save_measure_features, "Save the current table in a .csv file" )
        featlayout.addWidget(self.save_table)
        
        self.feature_group.setLayout(featlayout)
        self.feature_group.hide()
        all_layout.addWidget(self.feature_group)

        ## Track features
        self.trackfeat_group = QGroupBox(self.output_options[4])
        trackfeatlayout = QVBoxLayout()
        self.trackfeat_table = wid.add_button( "Track features table", self.show_trackfeature_table, "Measure track-related feature and show a table by track" )
        trackfeatlayout.addWidget(self.trackfeat_table)
        self.trackTable = FeaturesTable(self.viewer, self.epicure)
        trackfeatlayout.addWidget(self.trackTable)
        
        self.trackfeat_group.setLayout(trackfeatlayout)
        self.trackfeat_group.hide()
        all_layout.addWidget(self.trackfeat_group)

        ## Option to export events (Fiji ROI or table), + graphs ?
        self.export_event_group = QGroupBox(self.output_options[5])
        elayout = QVBoxLayout()
        self.choose_events_btn = wid.add_button( "Choose events", self.choose_events, "Open a window to select the events to export/measure" )
        elayout.addWidget( self.choose_events_btn )
        save_evt_line, self.save_evt_choice = wid.button_list( "Export events as", self.export_events, "Save the checked events as Fiji ROIs or .csv table" ) 
        self.save_evt_choice.addItem( "Fiji ROI" )
        elayout.addLayout( save_evt_line )

        self.export_event_group.setLayout( elayout )
        self.export_event_group.hide()
        all_layout.addWidget( self.export_event_group )
        
        
        ## Finished
        self.setLayout(all_layout)
        self.show_output_option()
        #self.setStyleSheet('QGroupBox {color: grey; background-color: rgb(35,45,50)} ')

    def get_current_settings( self ):
        """ Returns current settings of the widget """
        disp = {}
        disp["Apply on"] = self.output_mode.currentText() 
        disp["Current option"] = self.choose_output.currentText()
        disp = self.cell_features.get_current_settings( disp )
        disp = self.event_types.get_current_settings( disp )
        return disp

    def apply_settings( self, settings ):
        """ Set the current state of the widget from preferences if any """
        for setting, val in settings.items():
            if setting == "Apply on":
                self.output_mode.setCurrentText( val )
            if setting == "Current option":
                self.choose_output.setCurrentText( val )
            
        self.cell_features.apply_settings( settings )
        self.event_types.apply_settings( settings )

    def events_select( self, event, check ):
        """ Check/Uncheck the event in event types list """
        if event in self.event_types.types:
            self.event_types.types[ event ][0].setChecked( check )
        else:
            print(event+" not found in possible event types to export")

    def show_output_option(self):
        """ Show selected output panel """
        cur_option = self.choose_output.currentText()
        self.export_group.setVisible( cur_option == "Export to extern plugins" )
        self.export_seg_group.setVisible( cur_option == "Export segmentations" )
        self.feature_group.setVisible( cur_option == "Measure cell features" )
        self.trackfeat_group.setVisible( cur_option == "Measure track features" )
        self.export_event_group.setVisible( cur_option == "Export events" )

    def get_current_labels( self ):
        """ Get the cell labels to process according to current selection of apply on"""
        if self.output_mode.currentText() == "Only selected cell": 
            lab = self.epicure.seglayer.selected_label
            return [lab]
        if self.output_mode.currentText() == "All cells": 
            return self.epicure.get_labels()
        else:
            group = self.output_mode.currentText()
            label_group = self.epicure.groups[group]
            return label_group

            
    def get_selection_name(self):
        if self.output_mode.currentText() == "Only selected cell": 
            lab = self.epicure.seglayer.selected_label
            return "_cell_"+str(lab) 
        #if self.output_mode.currentText() == "Only checked cells":
        #    return "_checked_cells"
        if self.output_mode.currentText() == "All cells":
            return ""
        return "_"+self.output_mode.currentText()

    def save_measure_features(self):
        """ Save measures table to file whether it was created or not """
        if self.table is None or self.table_selection is None or self.selection_changed() :
            ut.show_warning("Create/update the table before")
            return
        outfile = self.epicure.outname()+"_features"+self.get_selection_name()+".xlsx"
        self.table.to_excel(outfile, sheet_name='EpiCureMeasures')
        if self.epicure.verbose > 0:
            ut.show_info("Measures saved in "+outfile)

    def save_one_roi(self, lab):
        """ Save the Rois of cell with label lab """
        keep = self.seglayer.data == lab
        rois = []
        if np.sum(keep) > 0:
            ## add 2D case
            for iframe, frame in enumerate(keep):
                if np.sum(frame) > 0:
                    contour = ut.get_contours(frame)
                    roi = self.create_roi(contour[0], iframe, lab)
                    rois.append(roi)

        roifile.roiwrite(self.epicure.outname()+"_rois_cell_"+str(lab)+".zip", rois, mode='w')

    def create_roi(self, contour, frame, label):
        croi = roifile.ImagejRoi()
        croi.version = 227
        croi.roitype = roifile.ROI_TYPE(0) ## polygon
        croi.n_coordinates = len(contour)
        croi.position = frame + 1
        croi.t_position = frame+1
        coords = []
        cent0 = 0
        cent1 = 0
        for cont in contour:
            coords.append([int(cont[1]), int(cont[0])])
            cent0 += cont[1]
            cent1 += cont[0]
        croi.integer_coordinates = np.array(coords)
        #croi.top = int(np.min(coords[0]))
        #croi.left = int(np.min(coords[1]))
        croi.name = str(frame+1).zfill(4)+'-'+str(int(cent0/len(contour))).zfill(4)+"-"+str(int(cent1/len(contour))).zfill(4)
        return croi
    
    def save_segmentation( self ):
        """ Save current segmentation in selected format """
        if self.output_mode.currentText() == "Only selected cell": 
            ## output only the selected cell
            lab = self.seglayer.selected_label
            if self.save_choice.currentText() == "ROI":
                self.save_one_roi(lab)
                if self.epicure.verbose > 0:
                    ut.show_info("Cell "+str(lab)+" saved to Fiji ROI")
                return
            else:
                tosave = np.zeros(self.seglayer.data.shape, dtype=self.epicure.dtype)
                if np.sum(self.seglayer.data==lab) > 0:
                    tosave[self.seglayer.data==lab] = lab
                endname = "_"+self.save_choice.currentText()+"_"+str(lab)+".tif"
        else:
            ## output all cells
            if self.output_mode.currentText() == "All cells":
                if self.save_choice.currentText() == "ROI":
                    self.save_all_rois()
                    return
                tosave = self.seglayer.data
                endname = "_"+self.save_choice.currentText()+".tif"
            else:
                ## or output only selected group
                group = self.output_mode.currentText()
                label_group = self.epicure.groups[group]
                if self.save_choice.currentText() == "ROI":
                    ncells = 0
                    for lab in label_group:
                        self.save_one_roi(lab)
                        ncells += 1
                    if self.epicure.verbose > 0:
                        ut.show_info(str(ncells)+" cells saved to Fiji ROIs")
                    return
                tosave = np.zeros(self.seglayer.data.shape, dtype=self.epicure.dtype)
                endname = "_"+self.save_choice.currentText()+"_"+self.output_mode.currentText()+".tif"
                for lab in label_group:
                    tosave[self.seglayer.data==lab] = lab
        
        ## save filled image (for label or skeleton) to file
        outname = os.path.join( self.epicure.outdir, self.epicure.imgname+endname )
        if self.save_choice.currentText() == "skeleton":
            tosave = ut.get_skeleton( tosave, verbose=self.epicure.verbose )
            ut.writeTif( tosave, outname, self.epicure.scale_xy, 'uint8', what="Skeleton" )
        else:
            ut.writeTif(tosave, outname, self.epicure.scale_xy, 'float32', what="Segmentation")
                
    def save_all_rois( self ):
        """ Save all cells to ROI format """
        ncells = 0
        for lab in np.unique(self.epicure.seglayer.data):
            self.save_one_roi(lab)
            ncells += 1
        if self.epicure.verbose > 0:
            ut.show_info(str(ncells)+" cells saved to Fiji ROIs")

    def choose_features( self ):
        """ Pop-up widget to choose the features to measure """
        self.cell_features.choose()

    def measure_features(self):
        """ Measure features and put them to table """
        def intensity_junction_cytoplasm(regionmask, intensity):
            """ Measure the intensity only on the contour of regionmask """
            footprint = disk(radius=self.epicure.thickness)
            inside = binary_erosion(regionmask, footprint)
            inside_intensity = ut.mean_nonzero(intensity*inside)
            dil_regionmask = binary_dilation(regionmask, footprint)
            periph = dil_regionmask^inside
            periph_intensity = ut.mean_nonzero(intensity*(regionmask^inside))
            return inside_intensity, periph_intensity
        
        if self.epicure.verbose > 0:
            print("Measuring features")
        self.viewer.window._status_bar._toggle_activity_dock(True)
        start_time = time.time()
        if self.output_mode.currentText() == "Only selected cell": 
            meas = np.zeros(self.epicure.seglayer.data.shape, self.epicure.dtype)
            lab = self.epicure.seglayer.selected_label
            meas[self.epicure.seglayer.data==lab] = lab
        else:
            if self.output_mode.currentText() == "All cells": 
                meas = self.epicure.seglayer.data
            else:
                group = self.output_mode.currentText()
                meas = np.zeros(self.epicure.seglayer.data.shape, self.epicure.dtype)
                label_group = self.epicure.groups[group]
                for lab in label_group:
                    meas[self.epicure.seglayer.data==lab] = lab
            
        properties, other_features, int_feat, int_extrafeat = self.cell_features.get_features()
        do_channels = self.cell_features.get_channels()
        ## prepare intensity extra properties if necessary
        extra_prop = []
        if "intensity_junction_cytoplasm" in int_extrafeat:
            extra_prop = extra_prop + [intensity_junction_cytoplasm]

        extra_properties = []
        if (do_channels is not None) and ("Movie" in do_channels):
            properties = properties + int_feat
            for extra in int_extrafeat:
                if extra == "intensity_junction_cytoplasm":
                    extra_properties = extra_properties + [intensity_junction_cytoplasm]
        self.table = None
        for iframe, frame in progress(enumerate(meas)):
            frame_table = self.measure_one_frame( frame, properties, extra_properties, other_features, do_channels, int_feat, extra_prop, iframe )
            if self.table is None:
                self.table = pand.DataFrame(frame_table)
            else:
                self.table = pand.concat([self.table, pand.DataFrame(frame_table)])

        if "intensity_junction_cytoplasm" in self.table.keys():
            self.table = self.table.rename(columns={"intensity_junction_cytoplasm-0": "intensity_cytoplasm", "intensity_junction_cytoplasm-1":"intensity_junction"})
        self.table_selection = self.selection_choices.index(self.output_mode.currentText())
        self.viewer.window._status_bar._toggle_activity_dock(False)
        if self.epicure.verbose > 0:
            ut.show_info("Features measured in "+"{:.3f}".format((time.time()-start_time)/60)+" min")
        
    def measure_one_frame(self, img, properties, extra_properties, other_features, channels, int_feat, int_extrafeat, frame):
        """ Measure on one frame """
        if frame is not None:
            intimg = self.movlayer.data[frame]
        else:
            intimg = self.movlayer.data
        frame_table = ut.labels_table( img, intensity_image=intimg, properties=properties, extra_properties=extra_properties )
        ndata = len(frame_table["label"])
        if frame is not None:
            frame_table["frame"] = np.repeat(frame, ndata)

        ## add info of the cell group
        if "group" in other_features:
            frame_group = self.epicure.get_groups(list(frame_table["label"]), numeric=False)
            frame_table["group"] = frame_group

        ### Measure intensity features in other chanels if option is on
        if (channels is not None):
            for chan in channels:
                ## if it's movie, already measured in the general measure
                if chan == "Movie":
                    continue
                ## otherwise, do a new measure on the selected channels
                oimg = self.viewer.layers[chan].data
                if frame is not None:
                    intimg = oimg[frame]
                else:
                    intimg = oimg
                frame_tab = ut.labels_table( img, intensity_image=intimg, properties=int_feat, extra_properties=int_extrafeat )
                for add_prop in int_feat:
                    frame_table[add_prop+"_"+str(chan)] = frame_tab[add_prop]
                if "intensity_junction_cytoplasm-0" in frame_tab.keys():
                    frame_table["intensity_cytoplasm_"+chan] = frame_tab["intensity_junction_cytoplasm-0"]
                    frame_table["intensity_junction_"+str(chan)] = frame_tab["intensity_junction_cytoplasm-1"]

        
        ## add features of neighbors relationship with graph
        do_neighbor = "NbNeighbors" in other_features
        do_border = "Border" in other_features
        if do_neighbor or do_border:
            touchlab = expand_labels(img, distance=3)  ## be sure that labels touch
            graph = RAG( touchlab, connectivity=2)
            adj_bg = []
            if 0 in graph.nodes:
                adj_bg = list(graph.adj[0])
                graph.remove_node(0)
            
            if do_neighbor:
                frame_table["NbNeighbors"] = np.repeat(-1, ndata)
            if do_border:
                frame_table["Border"] = np.repeat(-1, ndata)
            nodes = list(graph.nodes)
            for label in nodes:
                nneighbor = len(graph.adj[label])
                outer = int( label in adj_bg )
                rlabel = (frame_table["label"] == label)
                if do_neighbor:
                    frame_table["NbNeighbors"][rlabel] = nneighbor
                if do_border:
                    frame_table["Border"][rlabel] = outer

        return frame_table

    def selection_changed(self):
        if self.table_selection is None:
            return True
        return self.output_mode.currentText() != self.selection_choices[self.table_selection]

    def update_selection_list(self):
        """ Update the possible selection from group cell list """
        self.selection_choices = ["Only selected cell", "All cells"]
        for group in self.epicure.groups.keys():
            self.selection_choices.append(group)
        self.output_mode.clear()
        for sel in self.selection_choices:
            self.output_mode.addItem(sel)

    def show_table(self):
        """ Show the measurement table """
        #disable automatic update (slow)
        #if self.table is None:
            ## create the table and connect action to update it automatically
            #self.output_mode.currentIndexChanged.connect(self.show_table)
            #self.measure_other_chanels_cbox.stateChanged.connect(self.show_table)
            #self.feature_graph_cbox.stateChanged.connect(self.show_table)
            #self.feature_intensity_cbox.stateChanged.connect(self.show_table)
            #self.feature_shape_cbox.stateChanged.connect(self.show_table)
        
        ut.set_active_layer( self.viewer, "Segmentation" )
        self.show_feature_map.clear()
        self.show_feature_map.addItem("")
        laynames = [lay.name for lay in self.viewer.layers]
        for lay in laynames:
            if lay.startswith("Map_"):
                ut.remove_layer(self.viewer, lay)
        self.measure_features()
        self.featTable.set_table(self.table)
        featlist = self.table.keys()
        for feat in featlist:
            self.show_feature_map.addItem(feat)
        self.temp_graph.setEnabled(True)
        if self.tplots is not None:
            self.tplots.update_table(self.table)

    def show_feature(self):
        """ Add the image map of the selected feature """
        feat = self.show_feature_map.currentText()
        if (feat is not None) and (feat != ""):
            if feat in self.table.keys():
                values = list(self.table[feat])
                if feat == "group":
                    for i, val in enumerate(values):
                        if (val is None) or (val == 'None'):
                            values[i] = 0
                        else:
                            values[i] = list(self.epicure.groups.keys()).index(val) + 1
                labels = list(self.table["label"])
                frames = None
                if "frame" in self.table:
                    frames = list(self.table["frame"])
                self.draw_map(labels, values, frames, feat)

    def draw_map(self, labels, values, frames, featname):
        """ Add image layer of values by label """
        ## special feature: orientation, draw the axis instead
        self.viewer.window._status_bar._toggle_activity_dock(True)
        mapfeat = np.empty(self.epicure.seg.shape, dtype="float16")
        mapfeat[:] = np.nan
        if frames is not None:
            for frame, lab, val in progress(zip(frames, labels, values)):
                cell = self.seglayer.data[frame]==lab
                (mapfeat[frame])[cell] = val
        else:
            for lab, val in progress(zip(labels, values)):
                cell = self.seglayer.data==lab
                mapfeat[cell] = val
        ut.remove_layer(self.viewer, "Map_"+featname)
        self.viewer.add_image(mapfeat, name="Map_"+featname)
        self.viewer.window._status_bar._toggle_activity_dock(False)

    def draw_orientation( self ):
        """ Display the cells orientation axis in a new layer """
        ## check that necessary features are measured
        ut.remove_layer( self.viewer, "CellOrientation" )
        feats = ["centroid-0", "centroid-1", "orientation"]
        if self.table is None:
            print("Features centroid and orientation necessary to draw orientation, but are not measured yet")
            return
        for feat in feats:
            if feat not in self.table.keys():
                print("Feature "+feat+" necessary to draw orientation, but was not measured")
                return
        ## ok, can work now
        self.viewer.window._status_bar._toggle_activity_dock(True)

        ## get the coordinates of the axis lines by getting the cell centroid, main orientation
        xs = np.array( self.table["centroid-0"] )
        ys = np.array( self.table["centroid-1"] )
        angles = np.array( self.table["orientation"] )
        lens = np.array( [10]*len(angles) )
        oriens = np.zeros( (self.epicure.seg.shape), dtype="uint8" )

        ## draw axis length depending on the eccentricity
        if "eccentricity" in self.table.keys():
            lens = np.array(self.table["eccentricity"]*16)             
        
        if "frame" in self.table:
            frames = np.array( self.table["frame"] ).astype(int)
        else:
            frames = np.array( [0]*len(angles) )

        ## draw the lines in between the two extreme points (using Shape layer is too slow on display for big movies)
        npts = 30
        xmax = oriens.shape[1]-1
        ymax = oriens.shape[2]-1
        for i in range(npts):
            xas = np.clip(xs - lens/2 * np.cos( angles ) * i/float(npts), 0, xmax).astype(int)
            xbs = np.clip(xs + lens/2 * np.cos( angles ) * i/float(npts), 0, xmax).astype(int)
            yas = np.clip(ys - lens/2 * np.sin( angles ) * i/float(npts), 0, ymax).astype(int)
            ybs = np.clip(ys + lens/2 * np.sin( angles ) * i/float(npts), 0, ymax).astype(int)
            oriens[ (frames, xas, yas) ] = 255
            oriens[ (frames, xbs, ybs) ] = 255
        
        self.viewer.add_image( oriens, name="CellOrientation", blending="additive", opacity=1 )
        self.viewer.window._status_bar._toggle_activity_dock(False)

    ################### Export to other plugins

    def to_griot(self):
        """ Export current frame to new viewer and makes it ready for Griotte plugin """
        try:
            from napari_griottes import make_graph
        except:
            ut.show_error("Plugin napari-griottes is not installed")
            return
        gview = napari.Viewer()
        tframe = ut.current_frame(self.viewer)
        segt = self.epicure.seglayer.data[tframe]
        touching_frame = self.touching_labels(segt)
        gview.add_labels(touching_frame, name="TouchingCells", opacity=1)
        gview.window.add_dock_widget(make_graph(), name="Griottes")

    def touching_labels(self, labs):
        """ Dilate labels so that they all touch """
        from skimage.segmentation import find_boundaries
        from skimage.morphology import skeletonize
        from skimage.morphology import binary_closing, binary_opening
        from skimage.segmentation import expand_labels
        if self.epicure.verbose > 0:
            print("********** Generate touching labels image ***********")

        ## skeletonize it
        skel = skeletonize( binary_closing( find_boundaries(labs), footprint=np.ones((10,10)) ) )
        ext = np.zeros(labs.shape, dtype="uint8")
        ext[labs==0] = 1
        ext = binary_opening(ext, footprint=np.ones((2,2)))
        newimg = expand_labels(labs, distance=4)
        newimg[ext>0] = 0
        return newimg
    
    def to_ncp(self):
        """ Export current frame to new viewer and makes it ready for napari-cluster-plots plugin """
        try:
            import napari_skimage_regionprops as nsr
        except:
            ut.show_error("Plugin napari-skimage-regionprops is not installed")
            return
        gview = napari.Viewer()
        tframe = ut.current_frame(self.viewer)
        segt = self.epicure.seglayer.data[tframe]
        moviet = self.epicure.viewer.layers["Movie"].data[tframe]
        lab = gview.add_labels(segt, name="Segmentation[t="+str(tframe)+"]", blending="additive")
        im = gview.add_image(moviet, name="Movie[t="+str(tframe)+"]", blending="additive")
        if self.epicure.verbose > 0:
            print("Measure features with napari-skimage-regionprops plugin...")
        nsr.regionprops_table(im.data, lab.data, size=True, intensity=True, perimeter=True, shape=True, position=True, moments=True, napari_viewer=gview)
        try:
            import napari_clusters_plotter as ncp
        except:
            ut.show_error("Plugin napari-clusters-plotter is not installed")
            return
        gview.window.add_dock_widget(ncp.ClusteringWidget(gview))
        gview.window.add_dock_widget(ncp.PlotterWidget(gview))

    ################### Temporal graphs

    def temporal_graphs(self):
        """ New window with temporal graph of the current table selection """
        #self.temporal_viewer = napari.Viewer()
        self.tplots = TemporalPlots( self.viewer, self.epicure )
        self.tplots.setTable(self.table)
        self.tplots.show()
        #self.plot_wid = self.viewer.window.add_dock_widget( self.tplots, name="Plots" )
        self.viewer.dims.events.current_step.connect(self.position_verticalline)
    
    def on_close_viewer(self):
        """ Temporal plots window is closed """
        if self.epicure.verbose > 1:
            print("Closed viewer")
        self.viewer.dims.events.current_step.disconnect(self.position_verticalline)
        self.temporal_viewer = None
        self.tplots = None

    def position_verticalline(self):
        """ Place the vertical line in the temporal graph to the current frame """
        #try:
        #    wid = self.tplots
        #except:
        #    self.on_close_viewer()
        if self.tplots is not None:
            self.tplots.move_framepos(self.viewer.dims.current_step[0])

    ############### track features 

    def show_trackfeature_table(self):
        """ Show the measurement of tracks table """
        self.measure_track_features()
        self.trackTable.set_table( self.table )
    
    def measure_track_features(self):
        """ Measure track features and put them to table """
        if self.epicure.verbose > 0:
            print("Measuring track features")
        self.viewer.window._status_bar._toggle_activity_dock(True)
        start_time = time.time()

        if self.output_mode.currentText() == "Only selected cell": 
            track_ids = self.epicure.seglayer.selected_label
        else:
            if self.output_mode.currentText() == "All cells": 
                track_ids = self.epicure.tracking.get_track_list()
            else:
                group = self.output_mode.currentText()
                track_ids = []
                label_group = self.epicure.groups[group]
                for lab in label_group:
                    track_ids.append(lab)
            
        properties = ["label", "area", "centroid"]
        self.table = None

        if type(track_ids) == int:
            track_ids = [track_ids]
        for itrack, track_id in progress(enumerate(track_ids)):
                track_frame = self.measure_one_track( track_id )
                if self.table is None:
                    self.table = pand.DataFrame([track_frame])
                else:
                    self.table = pand.concat([self.table, pand.DataFrame([track_frame])])

        self.table_selection = self.selection_choices.index(self.output_mode.currentText())
        self.viewer.window._status_bar._toggle_activity_dock(False)
        if self.epicure.verbose > 0:
            ut.show_info("Features measured in "+"{:.3f}".format((time.time()-start_time)/60)+" min")

    def measure_one_track( self, track_id ):
        """ Measure features of one track """
        track_features = self.epicure.tracking.measure_track_features( track_id )
        return track_features

    ############## Events functions

    def choose_events( self ):
        """ Pop-up widget to choose the event types to measure/export """
        self.event_types.choose()

    def export_events( self ):
        """ Export events of selected types """
        evt_types = self.event_types.get_types()
        export_type = self.save_evt_choice.currentText()
        if export_type == "Fiji ROI":
            ## keep only events related to selected cells
            labels = self.get_current_labels()
            if self.epicure.verbose > 2:
                print("Exporting events of type "+str(evt_types)+" to Fiji ROIs" )
            ## export each type of event in separate files
            for itype, evt_type in enumerate( evt_types ):
                evts = self.epicure.inspecting.get_events_from_type( evt_type )
                if len( evts ) > 0:
                    rois = []
                    for evt_sid in evts:
                        pos, label = self.epicure.inspecting.get_event_infos( evt_sid )
                        if label in labels:
                            roi = self.create_point_roi( pos, itype )
                            rois.append( roi )
                    outfile = self.epicure.outname()+"_rois_"+evt_type +""+self.get_selection_name()+".zip" 
                    roifile.roiwrite(outfile, rois, mode='w')
                    if self.epicure.verbose > 0:
                        print( "Events "+str( evt_type )+" saved in ROI file: "+outfile )
                else:
                    ## dont save anything if empty, just print info to user
                    if self.epicure.verbose > 0:
                        print( "No events of type "+str(evt_type)+"" )

    def create_point_roi( self, pos, cat=0 ):
        """ Create a point Fiji ROI """
        croi = roifile.ImagejRoi()
        croi.version = 227
        croi.roitype = roifile.ROI_TYPE(10)
        croi.name = str(pos[0]+1).zfill(4)+'-'+str(pos[1]).zfill(4)+"-"+str(pos[2]).zfill(4)
        croi.n_coordinates = 1
        croi.left = int(pos[2])
        croi.top = int(pos[1])
        croi.z_position = 1
        croi.t_position = pos[0]+1
        croi.c_position = 1
        croi.integer_coordinates = np.array( [[0,0]] )
        croi.stroke_width=3
        ncolors = 3
        if cat%ncolors == 0:  ## color type 0
            croi.stroke_color = b'\xff\x00\x00\xff'
        if cat%ncolors == 1:  ## color type 1
            croi.stroke_color = b'\xff\x00\xff\x00'
        if cat%ncolors == 2:  ## color type 2
            croi.stroke_color = b'\xff\xff\x00\x00'
        return croi


class CellFeatures(QWidget):
    """ Choice of features to measure """
    def __init__(self, chanlist):
        super().__init__()
        layout = QVBoxLayout()
        
        self.required = ["label"]
        self.features = {}
        self.chan_list = None
        
        other_list = ["group", "NbNeighbors", "Border"]
        feat_layout = self.add_feature_group( other_list, "other" )
        layout.addLayout( feat_layout )

        ## Add shape features
        shape_list = ["centroid", "area", "area_convex", "axis_major_length", "axis_minor_length", "feret_diameter_max", "equivalent_diameter_area", "eccentricity", "orientation", "perimeter", "solidity"]
        feat_layout = self.add_feature_group( shape_list, "prop" )
        layout.addLayout( feat_layout )

        int_lab = QLabel()
        int_lab.setText("Intensity features:")
        layout.addWidget( int_lab )
        intensity_list = ["intensity_mean", "intensity_min", "intensity_max"]
        extra_list = ["intensity_junction_cytoplasm"]
        feat_layout = self.add_feature_group( intensity_list, "intensity_prop" )
        layout.addLayout( feat_layout )
        feat_layout = self.add_feature_group( extra_list, "intensity_extra" )
        layout.addLayout( feat_layout )
        if len(chanlist) > 1:
            chan_lab = QLabel()
            chan_lab.setText("Measure intensity in channels:")
            layout.addWidget( chan_lab )
            self.chan_list = QListWidget()
            self.chan_list.addItems( chanlist )
            self.chan_list.setSelectionMode(aiv.MultiSelection)
            self.chan_list.item(0).setSelected(True)
            layout.addWidget( self.chan_list )

        bye = wid.add_button( "Ok", self.close, "Close the window" )
        layout.addWidget( bye )
        self.setLayout( layout )

    def add_feature_group( self, feat_list, feat_type ):
        """ Add features to the GUI """
        layout = QVBoxLayout()
        ncols = 3
        for i, feat in enumerate(feat_list):
            if i%ncols == 0:
                line = QHBoxLayout()
            feature_check = QCheckBox(text=""+feat)
            line.addWidget(feature_check)
            self.features[ feat ] = [feature_check, feat_type]
            feature_check.setChecked( True )
            if i%ncols == (ncols-1):
                layout.addLayout( line )
                line = None
        if line is not None:
            layout.addLayout( line )
        return layout


    def close( self ):
        """ Close the pop-up window """
        self.hide()

    def choose( self ):
        """ Show the interface to select the choices """
        self.show()

    def get_current_settings( self, setting ):
        """ Get current settings of check or not of features """
        for feat, feat_cbox in self.features.items():
            setting[feat] = feat_cbox[0].isChecked()
        return setting

    def apply_settings( self, settings ):
        """ Set the checkboxes from preferenced settings """
        for feat, checked in settings.items():
            if feat in self.features.keys():
                self.features[feat][0].setChecked( checked )
        
    def get_features( self ):
        """ Returns the list of features to measure """
        feats = self.required
        int_extra_feats = []
        int_feats = []
        other_feats = []
        self.do_intensity = False
        for feat, feat_cbox in self.features.items():
            if feat_cbox[0].isChecked():
                if feat_cbox[1] == "prop":
                    feats.append( feat )
                if feat_cbox[1] == "other":
                    other_feats.append( feat )
                if feat_cbox[1] == "intensity_prop":
                    int_feats.append( feat )
                    self.do_intensity = True
                if feat_cbox[1] == "intensity_extra":
                    int_extra_feats.append( feat )
                    self.do_intensity = True
        return feats, other_feats, int_feats, int_extra_feats

    def get_channels( self ):
        """ Returns the list of channels to measure """
        if self.do_intensity:
            if self.chan_list is not None:
                wid_channels = self.chan_list.selectedItems()
                channels = []
                for chan in wid_channels:
                    channels.append( chan.text() )
            else:
                channels = ["Movie"]
            return channels
        return None

class EventTypes(QWidget):
    """ Choice of event types to export/measure """
    def __init__(self):
        super().__init__()
        layout = QVBoxLayout()
        
        self.types = {}
        possible_types = [ "division", "suspect" ] 
        event_layout = self.add_events( possible_types )
        layout.addLayout( event_layout )

        bye = wid.add_button( "Ok", self.close, "Close the window" )
        layout.addWidget( bye )
        self.setLayout( layout )

    def add_events( self, event_list ):
        """ Add events to the GUI """
        layout = QVBoxLayout()
        ncols = 3
        for i, event in enumerate( event_list ):
            if i%ncols == 0:
                line = QHBoxLayout()
            event_check = QCheckBox(text=""+event)
            line.addWidget( event_check )
            self.types[ event ] = [ event_check ]
            event_check.setChecked( True )
            if i%ncols == (ncols-1):
                layout.addLayout( line )
                line = None
        if line is not None:
            layout.addLayout( line )
        return layout


    def close( self ):
        """ Close the pop-up window """
        self.hide()

    def choose( self ):
        """ Show the interface to select the choices """
        self.show()

    def get_current_settings( self, setting ):
        """ Get current settings of check or not of features """
        for event, event_cbox in self.types.items():
            setting[event] = event_cbox[0].isChecked()
        return setting

    def apply_settings( self, settings ):
        """ Set the checkboxes from preferenced settings """
        for evt, checked in settings.items():
            if evt in self.types.keys():
                self.types[evt][0].setChecked( checked )
        
    def get_types( self ):
        """ Returns the list of events to measure """
        events = []
        for evt, evt_cbox in self.types.items():
            if evt_cbox[0].isChecked():
                events.append( evt )
        return events

class FeaturesTable(QWidget):
    """ Widget to visualize and interact with the measurement table """

    def __init__(self, napari_viewer, epicure):
        super().__init__()
        self.viewer = napari_viewer
        self.epicure = epicure
        self.wid_table = QTableWidget()
        self.wid_table.setEditTriggers(QTableWidget.EditTrigger.NoEditTriggers)
        self.setLayout(QGridLayout())
        self.layout().addWidget(self.wid_table)
        self.wid_table.clicked.connect(self.show_label)
        self.wid_table.setSortingEnabled(True)

    def show_label(self):
        """ When click on the table, show selected cell """
        if self.wid_table is not None:
            row = self.wid_table.currentRow()
            self.epicure.seglayer.show_selected_label = False
            headers = [self.wid_table.horizontalHeaderItem(ind).text() for ind in range(self.wid_table.columnCount()) ]
            labelind = None
            if "label" in headers:
                labelind = headers.index("label") 
            if "Label" in headers:
                labelind = headers.index("Label") 
            frameind = None
            if "frame" in headers:
                frameind = headers.index("frame") 
            if labelind is not None and labelind >= 0:
                lab = int(self.wid_table.item(row, labelind).text())
                if frameind is not None:
                    ## set current frame to the selected row
                    frame = int(self.wid_table.item(row, frameind).text())
                    ut.set_frame(self.viewer, frame)
                else:
                    ## set current frame to the first frame where label or track is present
                    frame = self.epicure.tracking.get_first_frame( lab )
                    if frame is not None:
                        ut.set_frame(self.viewer, frame)
                self.epicure.seglayer.selected_label = lab
                self.epicure.seglayer.show_selected_label = True


    def get_features_list(self):
        """ Return list of measured features """
        return [ self.wid_table.horizontalHeaderItem(ind).text() for ind in range(self.wid_table.columnCount()) ]

    def set_table(self, table):
        self.wid_table.clear()
        self.wid_table.setRowCount(table.shape[0])
        self.wid_table.setColumnCount(table.shape[1])

        for c, column in enumerate(table.keys()):
            column_name = column
            self.wid_table.setHorizontalHeaderItem(c, QTableWidgetItem(column_name))
            for r, value in enumerate(table.get(column)):
                item = QTableWidgetItem()
                item.setData( Qt.EditRole, value)
                self.wid_table.setItem(r, c, item)

class TemporalPlots(QWidget):
    """ Widget to visualize and interact with temporal plots """

    def __init__(self, napari_viewer, epicure):
        super().__init__()
        self.viewer = napari_viewer
        self.epicure = epicure
        self.features_list = ["frame"]
        self.parameter_gui()
        self.vline = None
        self.ymin = None
        #self.viewer.window.add_dock_widget( self.plot_wid, name="Temporal plot" )
   
    def parameter_gui(self):
        """ add widget to choose plotting parameters """
        
        layout = QVBoxLayout()

        ## choice of feature to plot
        feat_choice, self.feature_choice = wid.list_line( label="Plot feature", descr="Choose the feature to plot", func=self.plot_feature )
        layout.addLayout(feat_choice)
        ## option to average by group
        self.avg_group = wid.add_check( "Average by groups", False, self.plot_feature, descr="Show a line by cell or a line by group" )
        layout.addWidget(self.avg_group)
        ## show the plot
        self.plot_wid = self.create_plotwidget()
        layout.addWidget(self.plot_wid)
        ## save plot or save data of the plot
        line = wid.double_button( "Save plot image", self.save_plot_image, "Save the grapic in a PNG file", "Save plot data", self.save_plot_data, "Save the value used for the plot in .csv file" )
        layout.addLayout( line )
        self.setLayout(layout)

    def setTable(self, table):
        """ Data table to plot """
        self.table = table
        self.features_list = self.table.keys()
        self.update_feature_list()

    def update_table(self, table):
        """ Update the current plot with the updated table """
        self.table = table
        curchoice = self.feature_choice.currentText()
        self.features_list = self.table.keys()
        self.update_feature_list()
        if curchoice in self.features_list:
            ind = list(self.features_list).index(curchoice)
            self.feature_choice.setCurrentIndex(ind)
        self.plot_feature()

    def update_feature_list(self):
        """ Update the list of feature in the GUI """
        self.feature_choice.clear()
        for feat in self.features_list:
            self.feature_choice.addItem(feat)
    
    def plot_feature(self):
        """ Plot the selected feature in the temporal graph """
        feat = self.feature_choice.currentText()
        if feat == "label":
            return
        if feat == "":
            return
        self.ax.cla()
        if "group" in self.table:
            tab = list(zip(self.table["frame"], self.table[feat], self.table["label"], self.table["group"]))
            self.df = pand.DataFrame( tab, columns=["frame", feat, "label", "group"] )
        else:
            tab = list(zip(self.table["frame"], self.table[feat], self.table["label"]))
            self.df = pand.DataFrame( tab, columns=["frame", feat, "label"] )
        self.df.set_index('frame', inplace=True)
        if "group" in self.table and self.avg_group.isChecked():
            self.dfmean = self.df.groupby(['group', 'frame'])[feat].mean().reset_index()
            self.dfmean.set_index('frame', inplace=True)
            self.df.columns.name = 'group'
            self.dfmean.groupby('group')[feat].plot(legend=False, ax=self.ax)
            self.ax.legend(np.unique(self.dfmean['group']))
        else:
            self.df.groupby('label')[feat].plot(legend=False, ax=self.ax)
        self.ax.set_ylabel(''+feat)
        self.ax.set_xlabel('Time (frame)')
        self.fig.canvas.draw_idle()
        self.ymin, self.ymax = self.ax.get_ylim()

    def save_plot_image( self ):
        """ Save current plot graphic to PNG image """
        feat = self.feature_choice.currentText()
        outfile = self.epicure.outname()+"_plot_"+feat+".png"
        if self.fig is not None:
            self.fig.savefig( outfile )
        if self.epicure.verbose > 0:
            ut.show_info("Measures saved in "+outfile)

    def save_plot_data( self ):
        """ Save the raw data to redraw the current plot to csv file """
        feat = self.feature_choice.currentText()
        outfile = self.epicure.outname()+"_time_"+feat+".csv"
        if self.avg_group.isChecked():
            data = self.dfmean.reset_index()[["frame", "group", feat]]
            data[["frame", "group", feat]].to_csv( outfile,  sep='\t', header=True, index=False )
        else:
            data = self.df.reset_index()[["frame", "label", feat]]
            data[["frame", "label", feat]].to_csv( outfile,  sep='\t', header=True, index=False )

    def move_framepos(self, frame):
        """ Move the vertical line showing the current frame position in the main window """
        if self.ax is not None:
            if self.ymin is None:
                self.ymin, self.ymax = self.ax.get_ylim()
            if self.vline is not None:
                self.vline.remove()
            ymin = float(self.ymin*1.01)
            ymax = float(self.ymax*0.99)
            self.vline = self.ax.vlines(x=frame, ymin=ymin, ymax=ymax, ls=':', color="0.6")
            self.fig.canvas.draw_idle()

    def create_plotwidget(self):
        """ Create plot window """
        mpl_widget = FigureCanvas( Figure(figsize=(6,6) ) )
        self.fig = mpl_widget.figure
        self.ax = mpl_widget.figure.subplots()
        return mpl_widget

    




import numpy as np
from math import ceil
from qtpy.QtWidgets import QHBoxLayout, QVBoxLayout, QWidget, QComboBox
from qtpy.QtCore import Qt
from magicgui.widgets import TextEdit
import epicure.Utils as ut
import epicure.epiwidgets as wid


class Displaying(QWidget):
    """ Propose some visualization options """

    def __init__(self, napari_viewer, epic):
        """ Create displaying widget instance """
        super().__init__()
        self.viewer = napari_viewer
        self.epicure = epic
        self.seglayer = self.viewer.layers["Segmentation"]
        self.gmode = 0  ## view only movie mode on/off
        self.dmode = 0  ## view with light segmentation on/off
        self.grid_color = [0.6, 0.7, 0.7, 0.7]  ## default grid color

        layout = QVBoxLayout()
        
        ## Show a text window with some summary of the file
        show_summary = wid.add_button( "Show summary", self.show_summary_window, "Pops-up a summary of the movie and segmentation informations" )
        layout.addWidget(show_summary)
        
        ## Option show segmentation skeleton
        self.show_skeleton = wid.add_check( "Show segmentation skeleton", False, self.show_skeleton_segmentation, "Add a layer with the segmentation skeleton (not automatically updated)" )
        layout.addWidget(self.show_skeleton)
        
        ## Option to show the movie and seg side by side
        show_sides = QHBoxLayout()
        self.show_side = wid.add_check( "Side by side view", False, self.show_side_side, "View the movie and the other layers side by side" )
        show_sides.addWidget( self.show_side )
        self.directions = QComboBox()
        self.directions.addItem( "Horizontal" )
        self.directions.addItem( "Vertical" )
        show_sides.addWidget( self.directions )
        self.directions.currentIndexChanged.connect( self.show_side_side )
        layout.addLayout( show_sides )
        
        ## Option show shifted segmentation
        self.show_shifted = wid.add_check( "Overlay previous segmentation", False, self.show_shifted_segmentation, "Overlay the (frame-1) segmentation on the current segmentation")
        layout.addWidget(self.show_shifted)
        
        ## Option show shifted movie (previous or next)
        show_prevmovie_line = QHBoxLayout()
        self.show_previous_movie = wid.add_check( "Overlay previous movie", False, self.show_shifted_previous_movie, "Overlay the (frame-1) of the movie on the current movie" )
        layout.addWidget(self.show_previous_movie)
        self.show_next_movie = wid.add_check( "Overlay next movie", False, self.show_shifted_next_movie, "Overlay (frame+1) of the movie on the current frame" )
        layout.addWidget(self.show_next_movie)
        
        ## Option create/show grid
        grid_line, self.show_grid_options, self.group_grid = wid.checkgroup_help( "Grid options", True, "Show/hide subpanel to control grid view", "Display#grid-options", self.epicure.display_colors, groupnb="group" )
        self.grid_parameters()
        layout.addWidget(self.show_grid_options)
        layout.addWidget(self.group_grid)
        
        save_pref = wid.add_button( "Set current settings as default", self.save_current_display, "Save the current settings so that EpiCure will open in the same state next time" )
        layout.addWidget(save_pref)
        
        self.add_display_overlay_message()
        self.key_bindings()        ## activate shortcuts for display options
        self.setLayout(layout)
        ut.set_active_layer( self.viewer, "Segmentation" )
    
    ### set current display as defaut
    def save_current_display( self ):
        """ Set current display parameters as defaut display """
        self.epicure.update_settings()
        self.epicure.pref.save()

    def get_current_settings( self ):
        """ Returns current display settings """
        disp = {}
        disp["Layers"] = {}
        for layer in self.viewer.layers:
            disp["Layers"][layer.name] = layer.visible
        disp["Show Grid"] = self.show_grid_options.isChecked()
        disp["Grid nrows"] = self.nrows.text()
        disp["Grid ncols"] = self.ncols.text()
        disp["Grid width"] = self.gwidth.text()
        disp["Grid color"] = self.grid_color
        disp["Show side on"] = self.show_side.isChecked()
        disp["Side direction"] = self.directions.currentText()
        if "EpicGrid" in self.viewer.layers:
            disp["Grid text"] = self.viewer.layers["EpicGrid"].text.visible
            disp["Grid color"] = self.viewer.layers["EpicGrid"].edge_color[0]
        return disp
    
    def apply_settings( self, settings ):
        """ Set current display to prefered settings """
        add_grid = False
        show_text = False
        ## read the settings and apply them
        for setty, val in settings.items():
            if setty == "Layers":
                for layname, layvis in val.items():
                    if layname in self.viewer.layers:
                        self.viewer.layers[layname].visible = layvis
                    else:
                        if layname == "EpicGrid":
                            add_grid = layvis 
                continue
            if setty == "Show Grid":
                self.show_grid_options.setChecked( val )
                continue
            if setty == "Grid nrows":
                self.nrows.setText( val )
                continue
            if setty == "Grid ncols":
                self.ncols.setText( val )
                continue
            if setty == "Grid width":
                self.gwidth.setText( val )
                continue
            if setty == "Grid text":
                show_text = val
                continue
            if setty == "Grid color":
                self.grid_color = val
                continue
            if setty == "Show side on":
                self.show_side.setChecked( val )
            if setty == "Side direction":
                self.directions.setCurrentText( val )
            
        ## if grid should be added, do it at the end when all values are updated
        if add_grid:
            self.add_grid()
            self.viewer.layers["EpicGrid"].text.visible = show_text



    ######### overlay message
    def add_display_overlay_message(self):
        """ Shortcut list for display options """
        disptext = "--- Display options --- \n"
        sdisp = self.epicure.shortcuts["Display"]
        disptext += ut.print_shortcuts( sdisp )
        self.epicure.overtext["Display"] = disptext

    def show_summary_window(self):
        """ Show a text window with some infos """
        summwin = TextEdit()
        summwin.name = "Epicure summary"
        summwin.value = self.epicure.get_summary()
        summwin.show()


    ################  Key binging for display options
    def key_bindings(self):
        sdisp = self.epicure.shortcuts["Display"]
        
        @self.seglayer.bind_key( sdisp["vis. segmentation"]["key"], overwrite=True )
        def see_segmentlayer(seglayer):
            seglayer.visible = not seglayer.visible
        
        @self.seglayer.bind_key( sdisp["vis. movie"]["key"], overwrite=True )
        def see_movielayer(seglayer):
            ut.inv_visibility(self.viewer, "Movie")
        
        @self.seglayer.bind_key( sdisp["vis. event"]["key"], overwrite=True )
        def see_eventslayer(seglayer):
            evlayer = self.viewer.layers["Events"]
            evlayer.visible = not evlayer.visible
        
        @self.seglayer.bind_key( sdisp["skeleton"]["key"], overwrite=True )
        def show_skeleton(seglayer):
            """ On/Off show skeleton """
            if self.show_skeleton.isChecked():
                self.show_skeleton.setChecked(False)
            else:
                self.show_skeleton.setChecked(True)
        
        @self.seglayer.bind_key( sdisp["show side"]["key"], overwrite=True )
        def show_byside(seglayer):
            self.show_side.setChecked( not self.show_side.isChecked() )
            self.show_side_side()

        @self.seglayer.bind_key( sdisp["increase"]["key"], overwrite=True )
        def contour_increase(seglayer):
            if seglayer is not None:
                seglayer.contour = seglayer.contour + 1
        
        @self.seglayer.bind_key( sdisp["decrease"]["key"], overwrite=True )
        def contour_decrease(seglayer):
            if seglayer is not None:
                if seglayer.contour > 0:
                    seglayer.contour = seglayer.contour - 1
        
        @self.seglayer.bind_key( sdisp["only movie"]["key"], overwrite=True )
        def see_onlymovielayer(seglayer):
            """ if in "g" mode, show only movie, else put back to previous views """
            if self.gmode == 0:
                self.lay_view = []
                for lay in self.viewer.layers:
                    self.lay_view.append( (lay, lay.visible) )
                    lay.visible = False
                ut.inv_visibility(self.viewer, "Movie")
                self.gmode = 1
            else:
                for lay, vis in self.lay_view:
                    lay.visible = vis
                self.gmode = 0

        @self.seglayer.bind_key( sdisp["light view"]["key"], overwrite=True )
        def segmentation_lightmode(seglayer):
            """ if in "d" mode, show only movie and light segmentation, else put back to previous views """
            if self.dmode == 0:
                self.light_view = []
                for lay in self.viewer.layers:
                    self.light_view.append( (lay, lay.visible) )
                    lay.visible = False
                ut.inv_visibility(self.viewer, "Movie")
                ut.inv_visibility(self.viewer, "Segmentation")
                self.unlight_contour = self.seglayer.contour
                self.unlight_opacity = self.seglayer.opacity
                self.seglayer.contour = 1
                self.seglayer.opacity = 0.2
                self.dmode = 1
            else:
                for lay, vis in self.light_view:
                    lay.visible = vis
                self.seglayer.contour = self.unlight_contour
                self.seglayer.opacity = self.unlight_opacity
                self.dmode = 0
        
        @self.seglayer.bind_key( sdisp["grid"]["key"], overwrite=True )
        def show_grid(seglayer):
            """ show/hide the grid to have a repere in space """
            self.show_grid()

    ### Display options
    def show_skeleton_segmentation(self):
        """ Show/hide/update skeleton """
        if "Skeleton" in self.viewer.layers:
            ut.remove_layer(self.viewer, "Skeleton")
        if self.show_skeleton.isChecked():
            self.epicure.add_skeleton()
            ut.set_active_layer( self.viewer, "Segmentation" )

    def show_side_side( self ):
        """ Show the layers side by side """
        layout_grid = self.viewer.grid
        if self.show_side.isChecked():
            stride =  len( self.viewer.layers ) - 1
            layout_grid.stride = stride
            layout_grid.shape = (2,1)
            if self.directions.currentText() == "Horizontal":
                layout_grid.shape = (1,2)
            layout_grid.enabled = True
        else:
            layout_grid.enabled = False


    def show_shifted_segmentation(self):
        """ Show/Hide temporally shifted segmentation on top of current one """
        if ("PrevSegmentation" in self.viewer.layers):
            if (not self.show_shifted.isChecked()):
                ut.remove_layer(self.viewer, "PrevSegmentation")
            else:
                lay = self.viewer.layers["PrevSegmentation"]
                lay.refresh()

        if ("PrevSegmentation" not in self.viewer.layers) and (self.show_shifted.isChecked()):
            if self.epicure.nframes > 1:
                layer = self.viewer.add_labels( self.seglayer.data, name="PrevSegmentation", blending="additive", opacity=0.4 )
                layer.contour = 2
                layer.translate = [1, 0, 0]
                self.seglayer.contour = 2
                self.seglayer.opacity = 0.6
            else:
                ut.show_warning("Still image, cannot show previous frames")

        ut.set_active_layer( self.viewer, "Segmentation" )
    
    def show_shifted_previous_movie(self):
        """ Show/Hide temporally shifted movie previous frame on top of current one """
        self.show_shifted_movie("PrevMovie", "red", 1)
    
    def show_shifted_next_movie(self):
        """ Show/Hide temporally shifted movie next frame on top of current one """
        self.show_shifted_movie("NextMovie", "green", -1)
    
    def show_shifted_movie(self, layname, color, translation):
        """ Show/Hide temporally shifted movie on top of current one """
        if (layname in self.viewer.layers):
            if (not self.show_previous_movie.isChecked()):
                ut.remove_layer(self.viewer, layname)
            else:
                lay = self.viewer.layers[layname]
                lay.refresh()

        if (layname not in self.viewer.layers) and (self.show_previous_movie.isChecked()):
            if self.epicure.nframes > 1:
                movlay = self.viewer.layers["Movie"]
                arr = movlay.data
                if translation == -1:
                    arr = movlay.data[1:,]
                layer = self.viewer.add_image( arr, name=layname, blending="additive", opacity=0.6, colormap=color )
                if translation == 1:
                    layer.translate = [translation, 0, 0]
                layer.contrast_limits=self.epicure.quantiles()
                layer.gamma=0.94
            else:
                ut.show_warning("Still image, cannot show previous frames")

        ut.set_active_layer( self.viewer, "Segmentation" )

    #### Show/load a grid to have a repere in space
    def grid_parameters(self):
        """ Interface to get grid parameters """
        grid_layout = QVBoxLayout()
        ## nrows
        rows_line, self.nrows = wid.value_line( "Nb rows", "4", "Number of rows of the grid" )
        grid_layout.addLayout(rows_line)
        ## ncols
        cols_line, self.ncols = wid.value_line( "Nb columns:", "4", "Number of columns in the grid" )
        grid_layout.addLayout(cols_line)
        ## grid edges width
        width_line, self.gwidth = wid.value_line( "Grid width:", "3", "Width of the grid displayed lines/columns" )
        grid_layout.addLayout(width_line)
       ## go for grid
        btn_add_grid = wid.add_button( "Add grid", self.add_grid, "Add a grid overlay to the main view" )
        grid_layout.addWidget(btn_add_grid)
        self.group_grid.setLayout(grid_layout)

    def add_grid(self):
        """ Create/Load a new grid and add it """
        ut.remove_layer(self.viewer, "EpicGrid")
        imshape = self.epicure.imgshape2D
        if imshape is None:
            ut.show_error("Load the image first")
            return
        nrows = int(self.nrows.text())
        ncols = int(self.ncols.text())
        wid = ceil(imshape[0]/nrows)
        hei = ceil(imshape[1]/ncols)
        rects = []
        rects_names = []
        gwidth = int(self.gwidth.text())
        for x in range(nrows):
            for y in range(ncols):
                rect = np.array([[x*wid, y*hei], [(x+1)*wid, (y+1)*hei]])
                rects.append(rect)
                rects_names.append(chr(65+x)+"_"+str(y))
        self.viewer.add_shapes(rects, name="EpicGrid", text=rects_names, face_color=[1,0,0,0], edge_color=self.grid_color, edge_width=gwidth, opacity=0.7)
        self.viewer.layers["EpicGrid"].text.visible = False
        ut.set_active_layer( self.viewer, "Segmentation" )

    def show_grid(self):
        """ Interface to create/load a grid for repere """
        if "EpicGrid" not in self.viewer.layers:
            self.add_grid()
        else:
            gridlay = self.viewer.layers["EpicGrid"]
            gridlay.visible = not gridlay.visible
            gridlay.edge_color = self.grid_color
